# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
This module manages the UPF pseudopotentials in the local repository.
"""
import re

import aiida.orm.user
from aiida.orm.data.singlefile import SinglefileData
from aiida.common.utils import classproperty


UPFGROUP_TYPE = 'data.upf.family'

_upfversion_regexp = re.compile(
    r"""
    \s*<UPF\s+version\s*="
       (?P<version>.*)">
   """, re.VERBOSE)

_element_v1_regexp = re.compile(
    r"""
    ^
    \s*
    (?P<element_name>[a-zA-Z]{1,2})
    \s+
    Element
    \s*
    $
   """, re.VERBOSE)

_element_v2_regexp = re.compile(
    r"""
    \s*
    element\s*=\s*(?P<quote_symbol>['"])\s*
    (?P<element_name>[a-zA-Z]{1,2})\s*
    (?P=quote_symbol).*
   """, re.VERBOSE)


def get_pseudos_from_structure(structure, family_name):
    """
    Given a family name (a UpfFamily group in the DB) and a AiiDA
    structure, return a dictionary associating each kind name with its
    UpfData object.

    :raise MultipleObjectsError: if more than one UPF for the same element is
       found in the group.
    :raise NotExistent: if no UPF for an element in the group is
       found in the group.
    """
    from aiida.common.exceptions import NotExistent, MultipleObjectsError

    family_pseudos = {}
    family = UpfData.get_upf_group(family_name)
    for node in family.nodes:
        if isinstance(node, UpfData):
            if node.element in family_pseudos:
                raise MultipleObjectsError(
                    "More than one UPF for element {} found in "
                    "family {}".format(node.element, family_name))
            family_pseudos[node.element] = node

    pseudo_list = {}
    for kind in structure.kinds:
        symbol = kind.symbol
        try:
            pseudo_list[kind.name] = family_pseudos[symbol]
        except KeyError:
            raise NotExistent("No UPF for element {} found in family {}".format(
                symbol, family_name))

    return pseudo_list


def get_pseudos_dict(structure, family_name):
    """
    Get a dictionary of {kind: pseudo} for all the elements within the given
    structure using a the given pseudo family name.

    :param structure: The structure that will be used.
    :param family_name: the name of the group containing the pseudos
    """
    from collections import defaultdict

    # A dict {kind_name: pseudo_object}
    kind_pseudo_dict = get_pseudos_from_structure(structure, family_name)

    # We have to group the species by pseudo, I use the pseudo PK
    # pseudo_dict will just map PK->pseudo_object
    pseudo_dict = {}
    # Will contain a list of all species of the pseudo with given PK
    pseudo_species = defaultdict(list)

    for kindname, pseudo in kind_pseudo_dict.iteritems():
        pseudo_dict[pseudo.pk] = pseudo
        pseudo_species[pseudo.pk].append(kindname)

    pseudos = {}
    for pseudo_pk in pseudo_dict:
        pseudo = pseudo_dict[pseudo_pk]
        kinds = pseudo_species[pseudo_pk]
        for kind in kinds:
            pseudos[kind] = pseudo

    return pseudos


def upload_upf_family(folder, group_name, group_description,
                      stop_if_existing=True):
    """
    Upload a set of UPF files in a given group.

    :param folder: a path containing all UPF files to be added.
        Only files ending in .UPF (case-insensitive) are considered.
    :param group_name: the name of the group to create. If it exists and is
        non-empty, a UniquenessError is raised.
    :param group_description: a string to be set as the group description.
        Overwrites previous descriptions, if the group was existing.
    :param stop_if_existing: if True, check for the md5 of the files and,
        if the file already exists in the DB, raises a MultipleObjectsError.
        If False, simply adds the existing UPFData node to the group.
    """
    import os

    import aiida.common
    from aiida.common import aiidalogger
    from aiida.orm import Group
    from aiida.common.exceptions import UniquenessError, NotExistent
    from aiida.orm.backend import construct_backend
    from aiida.orm.querybuilder import QueryBuilder
    if not os.path.isdir(folder):
        raise ValueError("folder must be a directory")

    backend = construct_backend()

    # only files, and only those ending with .upf or .UPF;
    # go to the real file if it is a symlink
    files = [os.path.realpath(os.path.join(folder, i))
             for i in os.listdir(folder) if
             os.path.isfile(os.path.join(folder, i)) and
             i.lower().endswith('.upf')]

    nfiles = len(files)

    automatic_user = backend.users.get_automatic_user()
    try:
        group = Group.get(name=group_name, type_string=UPFGROUP_TYPE)
        group_created = False
    except NotExistent:
        group = Group(name=group_name, type_string=UPFGROUP_TYPE, user=automatic_user)
        group_created = True

    if group.user.email != automatic_user.email:
        raise UniquenessError("There is already a UpfFamily group with name {}"
                              ", but it belongs to user {}, therefore you "
                              "cannot modify it".format(group_name,
                                                        group.user.email))

    # Always update description, even if the group already existed
    group.description = group_description

    # NOTE: GROUP SAVED ONLY AFTER CHECKS OF UNICITY

    pseudo_and_created = []

    for f in files:
        md5sum = aiida.common.utils.md5_file(f)
        qb = QueryBuilder()
        qb.append(UpfData, filters={'attributes.md5':{'==':md5sum}})
        existing_upf = qb.first()

        #~ existing_upf = UpfData.query(dbattributes__key="md5",
                                     #~ dbattributes__tval=md5sum)

        if existing_upf is None:
            # return the upfdata instances, not stored
            pseudo, created = UpfData.get_or_create(f, use_first=True,
                                                    store_upf=False)
            # to check whether only one upf per element exists
            # NOTE: actually, created has the meaning of "to_be_created"
            pseudo_and_created.append((pseudo, created))
        else:
            if stop_if_existing:
                raise ValueError(
                        "A UPF with identical MD5 to "
                        " {} cannot be added with stop_if_existing"
                        "".format(f)
                    )
            existing_upf = existing_upf[0]
            pseudo_and_created.append((existing_upf, False))

    # check whether pseudo are unique per element
    elements = [(i[0].element, i[0].md5sum) for i in pseudo_and_created]
    # If group already exists, check also that I am not inserting more than
    # once the same element
    if not group_created:
        for aiida_n in group.nodes:
            # Skip non-pseudos
            if not isinstance(aiida_n, UpfData):
                continue
            elements.append((aiida_n.element, aiida_n.md5sum))

    elements = set(elements)  # Discard elements with the same MD5, that would
    # not be stored twice
    elements_names = [e[0] for e in elements]

    if not len(elements_names) == len(set(elements_names)):
        duplicates = set([x for x in elements_names
                          if elements_names.count(x) > 1])
        duplicates_string = ", ".join(i for i in duplicates)
        raise UniquenessError("More than one UPF found for the elements: " +
                              duplicates_string + ".")

        # At this point, save the group, if still unstored
    if group_created:
        group.store()

    # save the upf in the database, and add them to group
    for pseudo, created in pseudo_and_created:
        if created:
            pseudo.store()

            aiidalogger.debug("New node {} created for file {}".format(
                pseudo.uuid, pseudo.filename))
        else:
            aiidalogger.debug("Reusing node {} for file {}".format(
                pseudo.uuid, pseudo.filename))

    # Add elements to the group all togetehr
    group.add_nodes(pseudo for pseudo, created in pseudo_and_created)

    nuploaded = len([_ for _, created in pseudo_and_created if created])

    return nfiles, nuploaded


def parse_upf(fname, check_filename=True):
    """
    Try to get relevant information from the UPF. For the moment, only the
    element name. Note that even UPF v.2 cannot be parsed with the XML minidom!
    (e.g. due to the & characters in the human-readable section).

    If check_filename is True, raise a ParsingError exception if the filename
    does not start with the element name.
    """
    import os

    from aiida.common.exceptions import ParsingError
    from aiida.common import aiidalogger
    # TODO: move these data in a 'chemistry' module
    from aiida.orm.data.structure import _valid_symbols

    parsed_data = {}

    with open(fname) as f:
        first_line = f.readline().strip()
        match = _upfversion_regexp.match(first_line)
        if match:
            version = match.group('version')
            aiidalogger.debug("Version found: {} for file {}".format(
                version, fname))
        else:
            aiidalogger.debug("Assuming version 1 for file {}".format(fname))
            version = "1"

        parsed_data['version'] = version
        try:
            version_major = int(version.partition('.')[0])
        except ValueError:
            # If the version string does not start with a dot, fallback
            # to version 1
            aiidalogger.debug("Falling back to version 1 for file {}, "
                              "version string '{}' unrecognized".format(
                fname, version))
            version_major = 1

        element = None
        if version_major == 1:
            for l in f:
                match = _element_v1_regexp.match(l.strip())
                if match:
                    element = match.group('element_name')
                    break
        else:  # all versions > 1
            for l in f:
                match = _element_v2_regexp.match(l.strip())
                if match:
                    element = match.group('element_name')
                    break

        if element is None:
            raise ParsingError("Unable to find the element of UPF {}".format(
                fname))
        element = element.capitalize()
        if element not in _valid_symbols:
            raise ParsingError("Unknown element symbol {} for file {}".format(
                element, fname))
        if check_filename:
            if not os.path.basename(fname).lower().startswith(
                    element.lower()):
                raise ParsingError("Filename {0} was recognized for element "
                                   "{1}, but the filename does not start "
                                   "with {1}".format(fname, element))

        parsed_data['element'] = element

    return parsed_data


# def test_parser(folder):
#    import os
#    from aiida.common.exceptions import ParsingError
#
#    for fn in os.listdir(folder):
#        if os.path.isfile(fn) and fn.lower().endswith('.upf'):
#            try:
#                data = parse_upf(os.path.join(folder, fn))
#            except ParsingError as e:
#                print "~~~~~~~~~~~~~~~~~~~~ ERROR: %s" % e.message



class UpfData(SinglefileData):
    """
    Function not yet documented.
    """

    @classmethod
    def get_or_create(cls, filename, use_first=False, store_upf=True):
        """
        Pass the same parameter of the init; if a file with the same md5
        is found, that UpfData is returned.

        :param filename: an absolute filename on disk
        :param use_first: if False (default), raise an exception if more than \
                one potential is found.\
                If it is True, instead, use the first available pseudopotential.
        :param bool store_upf: If false, the UpfData objects are not stored in
                the database. default=True.
        :return (upf, created): where upf is the UpfData object, and create is either\
            True if the object was created, or False if the object was retrieved\
            from the DB.
        """
        import aiida.common.utils
        import os

        if not os.path.isabs(filename):
            raise ValueError("filename must be an absolute path")
        md5 = aiida.common.utils.md5_file(filename)

        pseudos = cls.from_md5(md5)
        if len(pseudos) == 0:
            if store_upf:
                instance = cls(file=filename).store()
                return (instance, True)
            else:
                instance = cls(file=filename)
                return (instance, True)
        else:
            if len(pseudos) > 1:
                if use_first:
                    return (pseudos[0], False)
                else:
                    raise ValueError("More than one copy of a pseudopotential "
                                     "with the same MD5 has been found in the "
                                     "DB. pks={}".format(
                        ",".join([str(i.pk) for i in pseudos])))
            else:
                return (pseudos[0], False)

    @classproperty
    def upffamily_type_string(cls):
        return UPFGROUP_TYPE

    def store(self, *args, **kwargs):
        """
        Store the node, reparsing the file so that the md5 and the element
        are correctly reset.
        """
        from aiida.common.exceptions import ParsingError, ValidationError
        import aiida.common.utils

        if self._to_be_stored is False:
            return self

        upf_abspath = self.get_file_abs_path()
        if not upf_abspath:
            raise ValidationError("No valid UPF was passed!")

        parsed_data = parse_upf(upf_abspath)
        md5sum = aiida.common.utils.md5_file(upf_abspath)

        try:
            element = parsed_data['element']
        except KeyError:
            raise ParsingError("No 'element' parsed in the UPF file {};"
                               " unable to store".format(self.filename))

        self._set_attr('element', str(element))
        self._set_attr('md5', md5sum)

        return super(UpfData, self).store(*args, **kwargs)


    @classmethod
    def from_md5(cls, md5):
        """
        Return a list of all UPF pseudopotentials that match a given MD5 hash.

        Note that the hash has to be stored in a _md5 attribute, otherwise
        the pseudo will not be found.
        """
        from aiida.orm.querybuilder import QueryBuilder
        qb = QueryBuilder()
        qb.append(cls, filters={'attributes.md5': {'==': md5}})
        return [_ for [_] in qb.all()]

    def set_file(self, filename):
        """
        I pre-parse the file to store the attributes.
        """
        from aiida.common.exceptions import ParsingError
        import aiida.common.utils

        parsed_data = parse_upf(filename)
        md5sum = aiida.common.utils.md5_file(filename)

        try:
            element = parsed_data['element']
        except KeyError:
            raise ParsingError("No 'element' parsed in the UPF file {};"
                               " unable to store".format(self.filename))

        super(UpfData, self).set_file(filename)

        self._set_attr('element', str(element))
        self._set_attr('md5', md5sum)

    def get_upf_family_names(self):
        """
        Get the list of all upf family names to which the pseudo belongs
        """
        from aiida.orm import Group

        return [_.name for _ in Group.query(nodes=self,
                                            type_string=self.upffamily_type_string)]

    @property
    def element(self):
        return self.get_attr('element', None)

    @property
    def md5sum(self):
        return self.get_attr('md5', None)

    def _validate(self):
        from aiida.common.exceptions import ValidationError, ParsingError
        import aiida.common.utils

        super(UpfData, self)._validate()

        upf_abspath = self.get_file_abs_path()
        if not upf_abspath:
            raise ValidationError("No valid UPF was passed!")

        try:
            parsed_data = parse_upf(upf_abspath)
        except ParsingError:
            raise ValidationError("The file '{}' could not be "
                                  "parsed".format(upf_abspath))
        md5 = aiida.common.utils.md5_file(upf_abspath)

        try:
            element = parsed_data['element']
        except KeyError:
            raise ValidationError("No 'element' could be parsed in the UPF "
                                  "file {}".format(upf_abspath))

        try:
            attr_element = self.get_attr('element')
        except AttributeError:
            raise ValidationError("attribute 'element' not set.")

        try:
            attr_md5 = self.get_attr('md5')
        except AttributeError:
            raise ValidationError("attribute 'md5' not set.")

        if attr_element != element:
            raise ValidationError("Attribute 'element' says '{}' but '{}' was "
                                  "parsed instead.".format(
                attr_element, element))

        if attr_md5 != md5:
            raise ValidationError("Attribute 'md5' says '{}' but '{}' was "
                                  "parsed instead.".format(
                attr_md5, md5))


    @classmethod
    def get_upf_group(cls, group_name):
        """
        Return the UpfFamily group with the given name.
        """
        from aiida.orm import Group

        return Group.get(name=group_name, type_string=cls.upffamily_type_string)


    @classmethod
    def get_upf_groups(cls, filter_elements=None, user=None):
        """
        Return all names of groups of type UpfFamily, possibly with some filters.

        :param filter_elements: A string or a list of strings.
               If present, returns only the groups that contains one Upf for
               every element present in the list. Default=None, meaning that
               all families are returned.
        :param user: if None (default), return the groups for all users.
               If defined, it should be either a DbUser instance, or a string
               for the username (that is, the user email).
        """
        from aiida.orm import Group

        group_query_params = {"type_string": cls.upffamily_type_string}

        if user is not None:
            group_query_params['user'] = user

        if isinstance(filter_elements, basestring):
            filter_elements = [filter_elements]

        if filter_elements is not None:
            actual_filter_elements = {_.capitalize() for _ in filter_elements}

            group_query_params['node_attributes'] = {
                'element': actual_filter_elements}

        all_upf_groups = Group.query(**group_query_params)

        groups = [(g.name, g) for g in all_upf_groups]
        # Sort by name
        groups.sort()
        # Return the groups, without name
        return [_[1] for _ in groups]




# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
import json

from aiida.backends.djsite.db.models import DbComputer, DbAuthInfo
from aiida.orm.authinfo import AuthInfo, AuthInfoCollection
from aiida.common import exceptions
from aiida.common.utils import type_check

from . import computer as computers
from . import user as users
from . import utils


class DjangoAuthInfoCollection(AuthInfoCollection):

    def create(self, computer, user):
        """
        Create a AuthInfo given a computer and a user

        :param computer: A Computer or DbComputer instance
        :param user: A User or DbUser instance
        :return: a AuthInfo object associated to the given computer and User
        """
        return DjangoAuthInfo(self, computer, user)

    def get(self, computer, user):
        """
        Return a AuthInfo given a computer and a user

        :param computer: A Computer or DbComputer instance
        :param user: A User or DbUser instance
        :return: a AuthInfo object associated to the given computer and User, if any
        :raise NotExistent: if the user is not configured to use computer
        :raise sqlalchemy.orm.exc.MultipleResultsFound: if the user is configured
             more than once to use the computer! Should never happen
        """
        from django.core.exceptions import (ObjectDoesNotExist,
                                            MultipleObjectsReturned)

        try:
            authinfo = DbAuthInfo.objects.get(
                # converts from name, Computer or DbComputer instance to
                # a DbComputer instance
                dbcomputer=DbComputer.get_dbcomputer(computer),
                aiidauser=user.id)

            return self._from_dbmodel(authinfo)
        except ObjectDoesNotExist:
            raise exceptions.NotExistent(
                "The aiida user {} is not configured to use computer {}".format(
                    user.email, computer.name))
        except MultipleObjectsReturned:
            raise exceptions.ConfigurationError(
                "The aiida user {} is configured more than once to use "
                "computer {}! Only one configuration is allowed".format(
                    user.email, computer.name))

    def _from_dbmodel(self, dbmodel):
        return DjangoAuthInfo._from_dbmodel(self, dbmodel)


class DjangoAuthInfo(AuthInfo):
    """
    AuthInfo implementation for Django
    """

    @classmethod
    def _from_dbmodel(cls, backend, dbmodel):
        type_check(dbmodel, DbAuthInfo)
        authinfo = cls.__new__(cls)
        super(DjangoAuthInfo, authinfo).__init__(backend)
        authinfo._dbauthinfo = utils.ModelWrapper(dbmodel)
        return authinfo

    def __init__(self, backend, computer, user):
        """
        Construct a DjangoAuthInfo

        """
        from aiida.orm.computer import Computer

        super(DjangoAuthInfo, self).__init__(backend)
        type_check(user, users.DjangoUser)

        # Takes care of always getting a Computer instance from a DbComputer, Computer or string
        dbcomputer = Computer.get(computer).dbcomputer

        self._dbauthinfo = utils.ModelWrapper(
            DbAuthInfo(dbcomputer=dbcomputer, aiidauser=user.dbuser))

    @property
    def dbauthinfo(self):
        return self._dbauthinfo._model

    @property
    def is_stored(self):
        """
        Is it already stored or not?

        :return: Boolean
        """
        return self._dbauthinfo.is_saved()

    @property
    def id(self):
        return self._dbauthinfo.id

    @property
    def enabled(self):
        return self._dbauthinfo.enabled

    @enabled.setter
    def enabled(self, value):
        self._dbauthinfo.enabled = value

    @property
    def computer(self):
        return computers.Computer.get(self._dbauthinfo.dbcomputer)

    @property
    def user(self):
        return self._backend.users._from_dbmodel(self._dbauthinfo.aiidauser)

    def get_auth_params(self):
        """
        Get the auth_params dictionary from the DB

        :return: a dictionary
        """
        try:
            return json.loads(self._dbauthinfo.auth_params)
        except ValueError:
            email = self._dbauthinfo.aiidauser.email
            hostname = self._dbauthinfo.dbcomputer.hostname
            raise exceptions.DbContentError(
                "Error while reading auth_params for dbauthinfo, aiidauser={}, computer={}".format(email, hostname))

    def set_auth_params(self, auth_params):
        """
        Replace the auth_params dictionary in the DB with the provided dictionary
        """
        import json

        # Raises ValueError if data is not JSON-serializable
        self._dbauthinfo.auth_params = json.dumps(auth_params)

    def get_metadata(self):
        """
        Get the metadata dictionary from the DB

        :return: a dictionary
        """
        import json

        try:
            return json.loads(self._dbauthinfo.metadata)
        except ValueError:
            raise exceptions.DbContentError(
                "Error while reading metadata for dbauthinfo, aiidauser={}, computer={}".format(
                    self.aiidauser.email, self.dbcomputer.hostname))

    def set_metadata(self, metadata):
        """
        Replace the metadata dictionary in the DB with the provided dictionary
        """
        # Raises ValueError if data is not JSON-serializable
        self._dbauthinfo.metadata = json.dumps(metadata)

    def store(self):
        """
        Store the AuthInfo (possibly updating values if changed)

        :return: the AuthInfo instance
        """
        self._dbauthinfo.save()
        return self



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

import abc

from aiida.common.hashing import is_password_usable
from aiida.common import exceptions

__all__ = ['User', 'UserCollection']


class UserCollection(object):
    """
    The collection of users stored in a backend
    """

    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def create(self, email):
        """
        Create a user with the provided email address

        :param email: An email address for the user
        :return: A new user object
        :rtype: :class:`User`
        """
        pass

    def find(self, email=None, id=None):
        """
        Final all users matching the given criteria
        :param email: An email address to search for
        :return: A collection of users matching the criteria
        """
        from .querybuilder import QueryBuilder

        qb = QueryBuilder()
        qb.append(User, filters={'email': {'==': email}})
        res = [_[0] for _ in qb.all()]
        if res is None:
            return []

        return res

    def get(self, email):
        results = self.find(email=email)
        if not results:
            raise exceptions.NotExistent()
        else:
            if len(results) > 1:
                raise exceptions.MultipleObjectsError()
            else:
                return results[0]

    def get_automatic_user(self):
        from aiida.common.utils import get_configured_user_email

        email = get_configured_user_email()
        if not email:
            return None

        try:
            return self.get(email)
        except (exceptions.MultipleObjectsError, exceptions.NotExistent):
            return None

    def all(self):
        """
        Final all users matching the given criteria
        :param email:
        :return: A collection of users matching the criteria
        """
        from .querybuilder import QueryBuilder

        qb = QueryBuilder()
        qb.append(User)
        return [_[0] for _ in qb.all()]


class User(object):
    """
    This is the base class for User information in AiiDA.  An implementing
    backend needs to provide a concrete version.
    """
    REQUIRED_FIELDS = ['first_name', 'last_name', 'institution']

    __metaclass__ = abc.ABCMeta

    def __init__(self, backend):
        self._backend = backend

    @property
    def backend(self):
        return self._backend

    @abc.abstractproperty
    def pk(self):
        pass

    @abc.abstractproperty
    def id(self):
        pass

    @abc.abstractproperty
    def is_stored(self):
        """
        Is the user stored

        :return: True if stored, False otherwise
        :rtype: bool
        """
        pass

    @abc.abstractmethod
    def store(self):
        pass

    @abc.abstractproperty
    def email(self):
        pass

    @abc.abstractmethod
    @email.setter
    def email(self, val):
        pass

    @property
    def password(self):
        return self._get_password()

    @password.setter
    def password(self, val):
        from aiida.common.hashing import create_unusable_pass, pwd_context

        if val is None:
            pass_hash = create_unusable_pass()
        else:
            pass_hash = pwd_context.encrypt(val)

        self._set_password(pass_hash)

    @abc.abstractmethod
    def _get_password(self):
        pass

    @abc.abstractmethod
    def _set_password(self, new_pass):
        pass

    @abc.abstractproperty
    def first_name(self):
        pass

    @abc.abstractmethod
    @first_name.setter
    def first_name(self, val):
        pass

    @abc.abstractproperty
    def last_name(self):
        pass

    @abc.abstractmethod
    @last_name.setter
    def last_name(self, val):
        pass

    @abc.abstractproperty
    def institution(self):
        pass

    @abc.abstractmethod
    @institution.setter
    def institution(self, val):
        pass

    @abc.abstractproperty
    def is_active(self):
        pass

    @abc.abstractmethod
    @is_active.setter
    def is_active(self, val):
        pass

    @abc.abstractproperty
    def last_login(self):
        pass

    @abc.abstractmethod
    @last_login.setter
    def last_login(self, val):
        pass

    @abc.abstractproperty
    def date_joined(self):
        pass

    @abc.abstractmethod
    @date_joined.setter
    def date_joined(self, val):
        pass

    def has_usable_password(self):
        return is_password_usable(self._get_password())

    def get_full_name(self):
        """
        Return the user full name

        :return: the user full name
        """
        if self.first_name and self.last_name:
            return "{} {} ({})".format(self.first_name, self.last_name,
                                       self.email)
        elif self.first_name:
            return "{} ({})".format(self.first_name, self.email)
        elif self.last_name:
            return "{} ({})".format(self.last_name, self.email)
        else:
            return "{}".format(self.email)

    def get_short_name(self):
        """
        Return the user short name (typically, this returns the email)

        :return: The short name
        """
        return self.email

    @staticmethod
    def get_schema():
        """
        Every node property contains:

          - display_name: display name of the property
          - help text: short help text of the property
          - is_foreign_key: is the property foreign key to other type of the node
          - type: type of the property. e.g. str, dict, int

        :return: schema of the user
        """
        return {
            "date_joined": {
                "display_name": "User since",
                "help_text": "Date and time of registration",
                "is_foreign_key": False,
                "type": "datetime.datetime"
            },
            "email": {
                "display_name": "email",
                "help_text": "e-mail of the user",
                "is_foreign_key": False,
                "type": "str"
            },
            "first_name": {
                "display_name": "First name",
                "help_text": "First name of the user",
                "is_foreign_key": False,
                "type": "str"
            },
            "id": {
                "display_name": "Id",
                "help_text": "Id of the object",
                "is_foreign_key": False,
                "type": "int"
            },
            "institution": {
                "display_name": "Institution",
                "help_text": "Affiliation of the user",
                "is_foreign_key": False,
                "type": "str"
            },
            "is_active": {
                "display_name": "Active",
                "help_text": "True(False) if the user is active(not)",
                "is_foreign_key": False,
                "type": "bool"
            },
            "last_login": {
                "display_name": "Last login",
                "help_text": "Date and time of the last login",
                "is_foreign_key": False,
                "type": "datetime.datetime"
            },
            "last_name": {
                "display_name": "Last name",
                "help_text": "Last name of the user",
                "is_foreign_key": False,
                "type": "str"
            }
        }



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from aiida.orm.data.array import ArrayData
from aiida.orm.data.orbital import OrbitalData 
from aiida.common.orbital import  Orbital
from aiida.common.exceptions import   ValidationError, NotExistent
import copy
import numpy as np
from aiida.orm import load_node
from aiida.orm.data.array.bands import BandsData


class ProjectionData(OrbitalData, ArrayData): 
    """
    A class to handle arrays of projected wavefunction data. That is projections
    of a orbitals, usually an atomic-hydrogen orbital, onto a
    given bloch wavefunction, the bloch wavefunction being indexed by
    s, n, and k. E.g. the elements are the projections described as
    < orbital | Bloch wavefunction (s,n,k) >
    """
    def _check_projections_bands(self, projection_array):
        """
        Checks to make sure that a reference bandsdata is already set, and that
        projection_array is of the same shape of the bands data

        :param projwfc_arrays: nk x nb x nwfc array, to be
                               checked against bands

        :raise: AttributeError if energy is not already set
        :raise: AttributeError if input_array is not of same shape as
                dos_energy
        """
        try:
            shape_bands = np.shape(self.get_reference_bandsdata())
        except AttributeError:
            raise ValidationError('Bands must be set first, then projwfc')
        # The [0:2] is so that each array, and not collection of arrays
        # is used to make the comparison
        if np.shape(projection_array) != shape_bands:
            raise AttributeError('These arrays are not the same shape as'
                                 ' the bands')
        return None

    def set_reference_bandsdata(self, value):
        """
        Sets a reference bandsdata, creates a uuid link between this data
        object and a bandsdata object, must be set before any projection arrays

        :param value: a BandsData instance, a uuid or a pk
        :raise: NotExistent if there was no BandsData associated with uuid or pk
        """
        if isinstance(value, BandsData):
            uuid = value.uuid 
        else:
            try:
                pk = int(value)
                bands = load_node(pk=pk, type=BandsData)
                uuid = bands.uuid
            except ValueError:
                uuid = str(value)
                try:
                    bands = load_node(uuid=uuid, type=BandsData)
                    uuid = bands.uuid
                except :
                    raise NotExistent("The value passed to "
                                      "set_reference_bandsdata was not "
                                      "associated to any bandsdata")
                    
        self._set_attr('reference_bandsdata_uuid', uuid)

    def get_reference_bandsdata(self):
        """
        Returns the reference BandsData, using the set uuid via
        set_reference_bandsdata

        :return: a BandsData instance
        :raise AttributeError: if the bandsdata has not been set yet
        :raise NotExistent: if the bandsdata uuid did not retrieve bandsdata
        """
        try:
            uuid = self.get_attr('reference_bandsdata_uuid')
        except AttributeError:
            raise AttributeError("BandsData has not been set for this instance")
        try:
            #bands = load_node(uuid=uuid, type=BandsData)
            bands = load_node(uuid=uuid) #TODO switch to above once type
                                         # has been implemented for load_node
        except NotExistent:
            raise NotExistent("The bands referenced to this class have not been "
                              "found in this database.")
        return bands

    def _find_orbitals_and_indices(self, **kwargs):
        """
        Finds all the orbitals and their indicies associated with kwargs
        essential for retrieving the other indexed array parameters

        :param kwargs: kwargs that can call orbitals as in get_orbitals()
        :return: retrieve_indexes, list of indicicies of orbitals corresponding
                 to the kwargs
        :return: all_orbitals, list of orbitals to which the indexes correspond
        """
        # index_and_orbitals = self._get_orbitals_and_index()
        index_and_orbitals = []
        selected_orbitals = self.get_orbitals(**kwargs)
        selected_orb_dicts = [orb.get_orbital_dict() for orb
                              in selected_orbitals]
        all_orbitals = self.get_orbitals()
        all_orb_dicts = [orb.get_orbital_dict() for orb in all_orbitals]
        retrieve_indices = [i for i in range(len(all_orb_dicts))
                            if all_orb_dicts[i] in selected_orb_dicts]
        return  retrieve_indices, all_orbitals

    def get_pdos(self, **kwargs):
        """
        Retrieves all the pdos arrays corresponding to the input kwargs

        :param kwargs: inputs describing the orbitals associated with the pdos
                       arrays
        :return: a list of tuples containing the orbital, energy array and pdos
                 array associated with all orbitals that correspond to kwargs

        """
        retrieve_indices, all_orbitals = self._find_orbitals_and_indices(**kwargs)
        out_list = [(all_orbitals[i],
                    self.get_array("pdos_{}".format(
                    self._from_index_to_arrayname(i))),
                    self.get_array("energy_{}".format(
                    self._from_index_to_arrayname(i))) )
                    for i in retrieve_indices]
        return out_list
    
    def get_projections(self, **kwargs):
        """
        Retrieves all the pdos arrays corresponding to the input kwargs

        :param kwargs: inputs describing the orbitals associated with the pdos
                       arrays
        :return: a list of tuples containing the orbital, and projection arrays
                 associated with all orbitals that correspond to kwargs

        """
        retrieve_indices, all_orbitals = self._find_orbitals_and_indices(**kwargs)
        out_list = [(all_orbitals[i],
                    self.get_array("proj_{}".format(
                    self._from_index_to_arrayname(i))))
                    for i in retrieve_indices]
        return out_list

    def _from_index_to_arrayname(self, index):
        """
        Used internally to determine the array names.
        """
        return "array_{}".format(index)
    
    def set_projectiondata(self,list_of_orbitals, list_of_projections=None,
                             list_of_energy=None, list_of_pdos=None,
                             tags = None, bands_check=True):
        """
        Stores the projwfc_array using the projwfc_label, after validating both.

        :param list_of_orbitals: list of orbitals, of class orbital data.
                                 They should be the ones up on which the
                                 projection array corresponds with.

        :param list_of_projections: list of arrays of projections of a atomic
                              wavefunctions onto bloch wavefunctions. Since the
                              projection is for every bloch wavefunction which
                              can be specified by its spin (if used), band, and
                              kpoint the dimensions must be
                              nspin x nbands x nkpoints for the projwfc array.
                              Or nbands x nkpoints if spin is not used.

        :param energy_axis: list of energy axis for the list_of_pdos

        :param list_of_pdos: a list of projected density of states for the
                             atomic wavefunctions, units in states/eV

        :param tags: A list of tags, not supported currently.

        :param bands_check: if false, skips checks of whether the bands has
                            been already set, and whether the sizes match. For
                            use in parsers, where the BandsData has not yet
                            been stored and therefore get_reference_bandsdata
                            cannot be called
        """
        def single_to_list(item):
            """
            Checks if the item is a list or tuple, and converts it to a list
            if it is not already a list or tuple

            :param item: an object which may or may not be a list or tuple
            :return: item_list: the input item unchanged if list or tuple and
                                [item] otherwise
            """
            if isinstance(item, (list, tuple)):
                return item
            else:
                return [item]
            
        def array_list_checker(array_list, array_name, orb_length):
            """
            Does basic checks over everything in the array_list. Makes sure that
            all the arrays are np.ndarray floats, that the length is same as
            required_length, raises exception using array_name if there is
            a failure
            """
            if not all([isinstance(_,np.ndarray) for _ in array_list]):
                raise ValidationError("{} was not composed "
                                      "entirely of ndarrays".format(array_name))
            if len(array_list) != orb_length:
                raise ValidationError("{} did not have the same length as the "
                                      "list of orbitals".format(array_name))
        
        ##############
        list_of_orbitals = single_to_list(list_of_orbitals)
        list_of_orbitals = copy.deepcopy(list_of_orbitals)

        # validates the input data
        if not list_of_pdos and not list_of_projections:
            raise ValidationError("Must set either pdos or projections")
        if bool(list_of_energy) != bool(list_of_pdos):
            raise ValidationError("list_of_pdos and list_of_energy must always "
                                  "be set together")

        orb_length = len(list_of_orbitals)
        
        # verifies and sets the orbital dicts
        list_of_orbital_dicts = []
        for i in range(len(list_of_orbitals)):
            this_orbital = list_of_orbitals[i]
            orbital_dict = this_orbital.get_orbital_dict()
            OrbitalClass = self._get_orbital_class_from_orbital_dict(
                           orbital_dict)
            test_orbital = OrbitalClass()
            try:
                test_orbital.set_orbital_dict(orbital_dict)
            except ValidationError:
                raise ValueError("The orbital with dict {} "
                                 "failed.".format(orbital_dict))
            list_of_orbital_dicts.append(test_orbital.get_orbital_dict())
        self._set_attr('orbital_dicts', list_of_orbital_dicts)

        # verifies and sets the projections
        if list_of_projections:
            list_of_projections = single_to_list(list_of_projections)
            array_list_checker(list_of_projections, "projections", orb_length)
            for i in range(len(list_of_projections)):
                this_projection = list_of_projections[i]
                array_name = self._from_index_to_arrayname(i)
                if bands_check:
                    self._check_projections_bands(this_projection)
                self.set_array("proj_{}".format(array_name), this_projection)

        # verifies and sets both pdos and energy
        if list_of_pdos:
            list_of_pdos = single_to_list(list_of_pdos)
            list_of_energy = single_to_list(list_of_energy)
            array_list_checker(list_of_pdos, "pdos", orb_length)
            array_list_checker(list_of_energy, "energy", orb_length)
            for i in range(len(list_of_pdos)):
                this_pdos = list_of_pdos[i]
                this_energy = list_of_energy[i]
                array_name = self._from_index_to_arrayname(i)
                if bands_check:
                    self._check_projections_bands(this_projection)
                self.set_array("pdos_{}".format(array_name), this_pdos)
                self.set_array("energy_{}".format(array_name), this_energy)

        # verifies and sets the tags
        if tags is not None:
            try:
                if len(tags) != len(list_of_orbitals):
                    raise ValidationError("must set as many tags as projections")
            except IndexError:
                return ValidationError("tags must be a list")

            if not all([isinstance(_,basestring) for _ in tags]):
                raise ValidationError("Tags must set a list of strings")
            self._set_attr('tags', tags)

    def set_orbitals(self, **kwargs):
        """
        This method is inherited from OrbitalData, but is blocked here.
        If used will raise a NotImplementedError
        """
        raise NotImplementedError("You cannot set orbitals using this class!"
                                  " This class is for setting orbitals and "
                                  " projections only!")



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from aiida.common.orbital import Orbital
from aiida.common.exceptions import ValidationError, InputValidationError
import copy


conversion_dict = {  'S':{'S':{'angular_momentum': 0, 'magnetic_number':0}
                              },
                    'P':{'PZ':{'angular_momentum': 1, 'magnetic_number':0},
                         'PX':{'angular_momentum': 1, 'magnetic_number':1},
                         'PY':{'angular_momentum': 1, 'magnetic_number':2},
                              },
                   'D':{'DZ2':{'angular_momentum': 2, 'magnetic_number':0},
                        'DXZ':{'angular_momentum': 2, 'magnetic_number':1},
                        'DYZ':{'angular_momentum': 2, 'magnetic_number':2},
                     'DX2-Y2':{'angular_momentum': 2, 'magnetic_number':3},
                        'DXY':{'angular_momentum': 2, 'magnetic_number':4},
                              },
                   'F':{'FZ3':{'angular_momentum': 3, 'magnetic_number':0},
                       'FZX2':{'angular_momentum': 3, 'magnetic_number':1},
                       'FYZ2':{'angular_momentum': 3, 'magnetic_number':2},
                  'FZ(X2-Y2)':{'angular_momentum': 3, 'magnetic_number':3},
                       'FXYZ':{'angular_momentum': 3, 'magnetic_number':4},
                 'FX(X2-3Y2)':{'angular_momentum': 3, 'magnetic_number':5},
                 'FY(3X2-Y2)':{'angular_momentum': 3, 'magnetic_number':6},
                              },
                 'SP':{'SP-1':{'angular_momentum':-1, 'magnetic_number':0},
                       'SP-2':{'angular_momentum':-1, 'magnetic_number':1},
                              },
               'SP2':{'SP2-1':{'angular_momentum':-2, 'magnetic_number':0},
                      'SP2-2':{'angular_momentum':-2, 'magnetic_number':1},
                      'SP2-3':{'angular_momentum':-2, 'magnetic_number':2},
                              },
               'SP3':{'SP3-1':{'angular_momentum':-3, 'magnetic_number':0},
                      'SP3-2':{'angular_momentum':-3, 'magnetic_number':1},
                      'SP3-3':{'angular_momentum':-3, 'magnetic_number':2},
                      'SP3-4':{'angular_momentum':-3, 'magnetic_number':3},
                              },
             'SP3D':{'SP3D-1':{'angular_momentum':-4, 'magnetic_number':0},
                     'SP3D-2':{'angular_momentum':-4, 'magnetic_number':1},
                     'SP3D-3':{'angular_momentum':-4, 'magnetic_number':2},
                     'SP3D-4':{'angular_momentum':-4, 'magnetic_number':3},
                     'SP3D-5':{'angular_momentum':-4, 'magnetic_number':4},
                              },
           'SP3D2':{'SP3D2-1':{'angular_momentum':-5, 'magnetic_number':0},
                    'SP3D2-2':{'angular_momentum':-5, 'magnetic_number':1},
                    'SP3D2-3':{'angular_momentum':-5, 'magnetic_number':2},
                    'SP3D2-4':{'angular_momentum':-5, 'magnetic_number':3},
                    'SP3D2-5':{'angular_momentum':-5, 'magnetic_number':4},
                    'SP3D2-6':{'angular_momentum':-5, 'magnetic_number':5},
                      }}

class RealhydrogenOrbital(Orbital):
    """
    Orbitals for hydrogen, largely follows the conventions used by wannier90
    Object to handle the generation of real hydrogen orbitals and their
    hybrids, has methods for producing s, p, d, f, and sp, sp2, sp3, sp3d,
    sp3d2 hybrids. This method does not deal with the cannonical hydrogen
    orbitals which contain imaginary components.

    The orbitals described here are chiefly concerned with the symmetric
    aspects of the oribitals without the context of space. Therefore
    diffusitivity, position and atomic labels should be handled in the
    OrbitalData class.

    Following the notation of table 3.1, 3.2 of Wannier90 user guide
    http://www.wannier.org/doc/user_guide.pdf
    A brief description of what is meant by each of these labels:

    :param radial_nodes: the number of radial nodes (or inflections) if no
                         radial nodes are supplied, defaults to 0
    :param angular_momentum: Angular quantum number, using real orbitals
    :param magnetic_number: Magnetic quantum number, using real orbitals
    :param spin: spin, up (1) down (-1) or unspecified (0)

    The conventions regarding L and M correpsond to those used in
    wannier90 for all L greater than 0 the orbital is not hyrbridized
    see table 3.1 and for L less than 0 the orbital is hybridized see
    table 3.2. M then indexes all the possible orbitals from 0 to 2L for
    L > 0 and from 0 to (-L) for L < 0.
    """

    _default_quantum_number_fields = ('angular_momentum',
                                      'magnetic_number',
                                      'spin',
                                      'radial_nodes')
    _optional_fields = (['kind_name'])

    _default_fields = tuple(list(Orbital._base_fields) +
                            list(_default_quantum_number_fields) +
                            list(_optional_fields))

    def __str__(self):
        orb_dict = self.get_orbital_dict()
        orb_name = self.get_name_from_quantum_numbers(
                        orb_dict['angular_momentum'],
                        magnetic_number=orb_dict['magnetic_number'])
        position_string = "{0:.1f},{0:.1f},{0:.1f}".format(
                                           orb_dict['position'][0],
                                           orb_dict['position'][1],
                                           orb_dict['position'][2])
        out_string = "r{} {} orbital '{}' @ {}".format(
                                           orb_dict['radial_nodes'],
                                           orb_name,
                                           orb_dict['kind_name'],
                                           position_string)
        return out_string

    def _quantum_number_validator(self, value,value_string, value_range):
        """
        Helper function, checks that value is indeed a integer, or None
        (in which case it returns None) and that the value falls within
        the input range. Returns the value if it passes and will raise
        an Exception otherwise.

        :param value: the value to be validated
        :param value_string: the label of the value
        :param value_range: range of values the value may fall
        :return value: the validated number
        """
        if value is None:
            return None
        if type(value) is not int:
            raise Exception('{} must always be an integer you provided {}'
            'instead'.format(value_string, value))
        if value < min(value_range) or value > max(value_range):
            raise Exception('You supplied {}  of value {} but it must be in'
            'the range {} instead'.format( value_string, value,  value_range))
        return value

    def _validate_keys(self, input_dict):
        """
        Validates the keys otherwise raise ValidationError. Does basic
        validation from the parent followed by validations for the
        quantum numbers. Raises exceptions should the input_dict fail the
        valiation or if it contains any unsupported keywords.

        :param input_dict: the dictionary of keys to be validated
        :return validated_dict: a validated dictionary
        """
        validated_dict = super(RealhydrogenOrbital,
                               self)._validate_keys(input_dict)
        # removes all validated items from input_dict
        input_dict = {x: input_dict[x] for x in input_dict
                      if x not in validated_dict}
        # get quantum numbers
        quantum_number_dict = {}
        for key in RealhydrogenOrbital._default_quantum_number_fields:
                quantum_number_dict[key] = input_dict.pop(key, None)
        # state lower and upper limits
        accepted_range = {}
        accepted_range["angular_momentum"] = [-5, 3]
        accepted_range["radial_nodes"] = [0, 2]
        accepted_range["spin"] = [-1, 1]
        l = quantum_number_dict["angular_momentum"]
        if l >= 0:
            accepted_range["magnetic_number"] = [0, 2*l]
        else:
            accepted_range["magnetic_number"] = [0, -l]
        # Here the tests with the limits defined above
        for key in RealhydrogenOrbital._default_quantum_number_fields:
            validated_number = self._quantum_number_validator(
                quantum_number_dict[key], key, accepted_range[key])
            validated_dict[key] = validated_number

        if validated_dict["angular_momentum"] is None:
            raise InputValidationError("Must supply angular_momentum")
        if validated_dict["magnetic_number"] is None:
            raise InputValidationError("Must supply magnetic_number")
        if validated_dict["radial_nodes"] is None:
            validated_dict["radial_nodes"] = 0
        try:
            self.get_name_from_quantum_numbers(
                validated_dict['angular_momentum'],
                magnetic_number=validated_dict['magnetic_number'])
        except InputValidationError:
            raise InputValidationError("Invalid angular momentum magnetic"
                                       " number combination.")

        # Finally checks optional fields
        KindName = input_dict.pop('kind_name', None)
        if KindName:
            if not isinstance(KindName, (basestring, None)):
                raise ValidationError('If kind_name is provided must be string')
            validated_dict['kind_name'] = KindName
        Kind_index = input_dict.pop('kind_index', None)
        if Kind_index:
            if not isinstance(Kind_index, (int, None)):
                raise ValidationError('If kind_index is provided must be int')
            validated_dict['kind_index'] = Kind_index

        if input_dict:
            raise ValidationError("Some unrecognized keys: {}".
                                  format(input_dict.keys()))
        return validated_dict

    @classmethod
    def get_name_from_quantum_numbers(cls, angular_momentum,
                                      magnetic_number=None):
        """
        Returns the orbital_name correponding to the angular_momentum alone,
        or to both angular_number with magnetic_number. For example using
        angular_momentum=1 and magnetic_number=1 will return "Px"
        """
        # importing a copy of the conversion_dict
        convert_ref = copy.deepcopy(conversion_dict)
        orbital_name = [x for x in convert_ref if
                any([convert_ref[x][y]['angular_momentum'] ==
                angular_momentum for y in convert_ref[x]])]
        if len(orbital_name) == 0:
            raise InputValidationError("No orbital name corresponding to the "
            "angular_momentum {} could be found".format(angular_momentum))
        if magnetic_number is not None:
            # finds angu
            orbital_name = orbital_name[0]
            orbital_name = [x for x in convert_ref[orbital_name] if
                            convert_ref[orbital_name][x]['magnetic_number']
                            == magnetic_number]

            if len(orbital_name) == 0:
                raise InputValidationError("No orbital name corresponding to "
                                           "the magnetic_number {} could be "
                                           "found".format(magnetic_number))
        return orbital_name[0]

    @classmethod
    def get_quantum_numbers_from_name(cls, name):
        """
        Returns all the angular and magnetic numbers corresponding to name.
        For example, using "P" as name will return all quantum numbers
        associated with a "P" orbital, while "Px" will return only one set
        of quantum numbers, the ones associated with "Px"
        """
        # importing a copy of the conversion_dict
        convert_ref = copy.deepcopy(conversion_dict)
        if not isinstance(name, basestring):
            raise InputValidationError('Input name must be a string')
        name = name.upper()
        list_of_dicts = [convert_ref[x][y] for x in convert_ref for y in
                         convert_ref[x] if (y == name or x == name)]
        if len(list_of_dicts) == 0:
            raise InputValidationError("Invalid choice of projection name")
        return list_of_dicts

#     @classmethod
#     def generate_orbitals(cls, name=None, spin=0, radial_nodes=0,
#                           site=None, **kwargs):
#         """
#         Return any and all orbitals which match to those used in the
#         conversion dictionary returns a list of real hydrogen orbitals
#         which match, or raises an error if an invalid orbital name was
#         supplied. If a site is supplied, the coordinates will be applied
#         from it.
#
#         :param name: a string corresponding to a projeciton name
#         :param site: if a site is passed, will use the position of that site
#         :param spin: the spin quantum number
#         :param radial_nodes: number of radial nodes
#         :kwargs: additional parameters specifying properties of the orbitals
#         :return list_of_classes: a list of orbitals produced from the input
#         """
#
#         # finding appropriate raw_dict arguments from _conversion_dict
#         if name:
#             list_of_dicts = self.get_quantum_numbers_from_name(name)
#         else:
#             list_of_dicts = [{}]
#
#         # appending site data if possible
#         site_dict = {}
#         if site is not None:
#             site_dict = cls._prep_raw_dict_keys_from_site(site)
#         list_of_classes = []
#         for this_dict in list_of_dicts:
#             this_dict['spin'] = spin
#             this_dict['radial_nodes'] = radial_nodes
#             this_dict.update(kwargs)
#             this_dict.update(site_dict)
#             c = cls()  # intialize and set it
#             c.set_orbital_dict(this_dict)
#             list_of_classes.append(c)
#         return list_of_classes

# # Can move these later, but for now I want to have these functions handy
# def cart2sph(x, y, z):
#     """
#     Converts cartesian to unit vector spherical coordinates
#
#     :param x, y, z: absolute cartesian coordinates
#     :return: [r, theta, phi] unit vector in spherical coordinates
#     """
#     import numpy as np
#     coord_cart = np.array([x, y, z])
#     unit_cart = coord_cart/np.linalg.norm(coord_cart)
#     x,y,z = unit_cart # for clarity
#     r = np.linalg.norm(unit_cart)  # should always be 1.0
#     theta = np.arctan2(y,x)
#     phi = np.arctan2(z,r)
#     # note that this method will not return the r value, which is not stored
#     return [theta, phi]

# def sph2cart(theta, phi):
#     """
#     Converts spherical coordinates r, theta, phi to unit vector in cartesian
#     coordinates. r is not accepted as input because only unit vectors are
#     considered here.
#
#     : theta, phi: spherical coordinates vector
#     :return: x, y, z unit vector  in cartesian absolute coordinates
#     """
#     import numpy as np
#     # code is done in a few more steps than necessary, to make logic explicit
#     r = 1
#     x = r*np.cos(theta)*np.sin(phi)
#     y = r*np.sin(theta)*np.sin(phi)
#     z = r*np.cos(theta)
#     x = x/r
#     y = y/r
#     z = z/r
#     return [x, y, z]




# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
This allows to hook-up the AiIDA built-in RESTful API.
Main advantage of doing this by means of a verdi command is that different
profiles can be selected at hook-up (-p flag).

"""

from aiida.cmdline.baseclass import VerdiCommand

class Restapi(VerdiCommand):
    """
    verdi command used to hook up the AiIDA REST API.
    Example Usage:

    verdi -p <profile_name> restapi --host 127.0.0.5 --port 6789
    --config-dir <location of the onfig.py file>

    """
    import os
    import aiida.restapi

    # Defaults defined at class level
    default_host = "127.0.0.1"
    default_port = "5000"
    default_config_dir = os.path.join(os.path.split(os.path.abspath(
        aiida.restapi.__file__))[0], 'common')
    parse_aiida_profile = False

    def run(self, *args):
        """
        Hook up the default RESTful API of AiiDA.
        args include port, host, config_file
        """
        from aiida.restapi.api import App, AiidaApi
        from aiida.restapi.run_api import run_api

        # Construct dparameter dictionary
        kwargs = dict(
            hookup=True,
            prog_name=self.get_full_command_name(),
            default_host=self.default_host,
            default_port=self.default_port,
            default_config=self.default_config_dir,
            parse_aiida_profile=self.parse_aiida_profile,
            catch_internal_server=True)

        # Invoke the runner
        run_api(App, AiidaApi, *args, **kwargs)

    def complete(self, subargs_idx, subargs):
        """
        No particular completion features required
        """
        return ""



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

import plumpy
import uuid

from aiida.backends.testbase import AiidaTestCase
from aiida.orm.data.int import Int
from aiida.work import runners, rmq, test_utils


class TestProcessControl(AiidaTestCase):
    """
    Test AiiDA's RabbitMQ functionalities.
    """

    def setUp(self):
        super(TestProcessControl, self).setUp()
        prefix = '{}.{}'.format(self.__class__.__name__, uuid.uuid4())
        rmq_config = rmq.get_rmq_config(prefix)

        # These two need to share a common event loop otherwise the first will never send
        # the message while the daemon is running listening to intercept
        self.runner = runners.Runner(
            rmq_config=rmq_config,
            rmq_submit=True,
            poll_interval=0.)
        self.daemon_runner = runners.DaemonRunner(
            rmq_config=rmq_config,
            rmq_submit=True,
            loop=self.runner.loop,
            poll_interval=0.)

    def tearDown(self):
        self.daemon_runner.close()
        self.runner.close()
        super(TestProcessControl, self).tearDown()

    def test_submit_simple(self):
        # Launch the process
        calc_node = self.runner.submit(test_utils.DummyProcess)
        self._wait_for_calc(calc_node)

        self.assertTrue(calc_node.is_finished_ok)
        self.assertEqual(calc_node.process_state.value, plumpy.ProcessState.FINISHED.value)

    def test_launch_with_inputs(self):
        a = Int(5)
        b = Int(10)

        calc_node = self.runner.submit(test_utils.AddProcess, a=a, b=b)
        self._wait_for_calc(calc_node)
        self.assertTrue(calc_node.is_finished_ok)
        self.assertEqual(calc_node.process_state.value, plumpy.ProcessState.FINISHED.value)

    def test_submit_bad_input(self):
        with self.assertRaises(ValueError):
            self.runner.submit(test_utils.AddProcess, a=Int(5))

    def test_exception_process(self):
        calc_node = self.runner.submit(test_utils.ExceptionProcess)
        self._wait_for_calc(calc_node)

        self.assertFalse(calc_node.is_finished_ok)
        self.assertEqual(calc_node.process_state.value, plumpy.ProcessState.EXCEPTED.value)

    def test_pause(self):
        """ Testing sending a pause message to the process """
        calc_node = self.runner.submit(test_utils.WaitProcess)
        future = self.runner.rmq.pause_process(calc_node.pk)
        result = self.runner.run_until_complete(future)
        self.assertTrue(result)

    def test_pause_play(self):
        """ Test sending a pause and then a play message """
        calc_node = self.runner.submit(test_utils.WaitProcess)
        future = self.runner.rmq.pause_process(calc_node.pk)
        result = self.runner.run_until_complete(future)
        self.assertTrue(result)

        future = self.runner.rmq.play_process(calc_node.pk)
        result = self.runner.run_until_complete(future)
        self.assertTrue(result)

    def test_kill(self):
        """ Test sending a kill message """
        calc_node = self.runner.submit(test_utils.WaitProcess)
        future = self.runner.rmq.kill_process(calc_node.pk, "Sorry, you have to go mate")
        result = self.runner.run_until_complete(future)
        # TODO: Check kill message
        self.assertTrue(result)

    def _wait_for_calc(self, calc_node, timeout=2.):
        future = self.runner.get_calculation_future(calc_node.pk)
        self.runner.run_until_complete(future)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from aiida.orm.backend import Backend

from . import authinfo
from . import log
from . import user


class DjangoBackend(Backend):
    def __init__(self):
        self._log = log.DjangoLog()
        self._users = user.DjangoUserCollection()
        self._authinfos = authinfo.DjangoAuthInfoCollection()

    @property
    def log(self):
        return self._log

    @property
    def users(self):
        return self._users

    @property
    def authinfos(self):
        return self._authinfos



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
import os
import sys
import unittest
from unittest import (
    TestSuite, defaultTestLoader as test_loader)

from aiida.backends import settings
from aiida.backends.tests import get_db_test_list
from aiida.common.exceptions import ConfigurationError, TestsNotAllowedError, InternalError
from aiida.common.utils import classproperty


def check_if_tests_can_run():
    """
    Check if the tests can run (i.e., if we are in a test profile).
    Otherwise, raise TestsNotAllowedError.
    """
    from aiida import settings as settings2
    from aiida.common.setup import TEST_KEYWORD

    base_repo_path = os.path.basename(
        os.path.normpath(settings2.REPOSITORY_PATH))
    if (not settings.AIIDADB_PROFILE.startswith(TEST_KEYWORD) or
            TEST_KEYWORD not in base_repo_path or
            not settings2.DBNAME.startswith(TEST_KEYWORD)):
        msg = [
            "A non-test profile was given for tests. Please note "
            "that the test profile should have test specific "
            "database name and test specific repository name.",
            "Given profile: {}".format(settings.AIIDADB_PROFILE),
            "Related repository path: {}".format(base_repo_path),
            "Related database name: {}".format(settings2.DBNAME)]
        raise TestsNotAllowedError("\n".join(msg))


class AiidaTestCase(unittest.TestCase):
    """
    This is the base class for AiiDA tests, independent of the backend.

    Internally it loads the AiidaTestImplementation subclass according to the current backend
    """
    _class_was_setup = False
    __backend_instance = None

    @classmethod
    def get_backend_class(cls):
        from aiida.backends.testimplbase import AiidaTestImplementation

        from aiida.backends.profile import BACKEND_SQLA, BACKEND_DJANGO
        # Freeze the __impl_class after the first run
        if not hasattr(cls, '__impl_class'):
            if settings.BACKEND == BACKEND_SQLA:
                from aiida.backends.sqlalchemy.tests.testbase import (
                    SqlAlchemyTests)
                cls.__impl_class = SqlAlchemyTests
            elif settings.BACKEND == BACKEND_DJANGO:
                from aiida.backends.djsite.db.testbase import DjangoTests
                cls.__impl_class = DjangoTests
            else:
                raise ConfigurationError("Unknown backend type")

            # Check that it is of the right class
            if not issubclass(cls.__impl_class, AiidaTestImplementation):
                raise InternalError(
                    "The AiiDA test implementation is not of type "
                    "{}, that is not a subclass of AiidaTestImplementation".format(
                        cls.__impl_class.__name__
                    )
                )

        return cls.__impl_class

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        from aiida.orm.backend import construct_backend

        # Note: this will raise an exception, that will be seen as a test
        # failure. To be safe, you should do the same check also in the tearDownClass
        # to avoid that it is run
        check_if_tests_can_run()

        cls.__backend_instance = cls.get_backend_class()()
        cls.__backend_instance.setUpClass_method(*args, **kwargs)
        cls.backend = construct_backend()

        cls._class_was_setup = True

    def setUp(self):
        self.__backend_instance.setUp_method()

    def tearDown(self):
        self.__backend_instance.tearDown_method()

    @classmethod
    def insert_data(cls):
        cls.__backend_instance.insert_data()

    @classmethod
    def clean_db(cls):

        # Note: this will raise an exception, that will be seen as a test
        # failure. To be safe, you should do the same check also in the tearDownClass
        # to avoid that it is run
        check_if_tests_can_run()

        from aiida.common.exceptions import InvalidOperation

        if not cls._class_was_setup:
            raise InvalidOperation("You cannot call clean_db before running the setUpClass")

        cls.__backend_instance.clean_db()

    @classproperty
    def computer(cls):
        return cls.__backend_instance.get_computer()

    @classproperty
    def user_email(cls):
        return cls.__backend_instance.get_user_email()

    @classmethod
    def tearDownClass(cls, *args, **kwargs):

        # Double check for double security to avoid to run the tearDown
        # if this is not a test profile
        check_if_tests_can_run()

        cls.__backend_instance.tearDownClass_method(*args, **kwargs)


def run_aiida_db_tests(tests_to_run, verbose=False):
    """
    Run all tests specified in tests_to_run.
    Return the list of test results.
    """
    # Empty test suite that will be populated
    test_suite = TestSuite()

    actually_run_tests = []
    num_tests_expected = 0

    # To avoid adding more than once the same test
    # (e.g. if you type both db and db.xxx)
    found_modulenames = set()

    for test in set(tests_to_run):
        try:
            modulenames = get_db_test_list()[test]
        except KeyError:
            if verbose:
                print >> sys.stderr, "Unknown DB test {}... skipping".format(
                    test)
            continue
        actually_run_tests.append(test)

        for modulename in modulenames:
            if modulename not in found_modulenames:
                try:
                    test_suite.addTest(test_loader.loadTestsFromName(modulename))
                except AttributeError as exception:
                    try:
                        import importlib
                        import traceback
                        importlib.import_module(modulename)
                    except ImportError as exception:
                        print >> sys.stderr, (
                            "[CRITICAL] The module '{}' has an import error and the tests cannot be run:\n{}"
                                .format(modulename, traceback.format_exc(exception))
                        )
                        sys.exit(1)
                found_modulenames.add(modulename)

        num_tests_expected = test_suite.countTestCases()

    if verbose:
        print >> sys.stderr, (
            "DB tests that will be run: {} (expecting {} tests)".format(
                ",".join(actually_run_tests), num_tests_expected))

    results = unittest.TextTestRunner(failfast=False).run(test_suite)

    if verbose:
        print "Run tests: {}".format(results.testsRun)

    return results



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
import plumpy
import aiida.common
from aiida.common import aiidalogger
from aiida.orm.workflow import Workflow
from aiida.orm import Code, Computer


logger = aiidalogger.getChild('WorkflowDemo')


class WorkflowDemo(Workflow):
    def __init__(self, **kwargs):
        super(WorkflowDemo, self).__init__(**kwargs)

    def generate_calc(self):
        from aiida.orm import Code, Computer, CalculationFactory
        from aiida.common.datastructures import calc_states

        CustomCalc = CalculationFactory('simpleplugins.templatereplacer')

        computer = Computer.get("localhost")

        calc = CustomCalc(computer=computer, withmpi=True)
        calc.set_resources({"num_machines": 1, "num_mpiprocs_per_machine": 1})
        calc.store()
        calc._set_state(calc_states.FINISHED)
        calc._set_process_state(plumpy.ProcessState.FINISHED)

        return calc

    @Workflow.step
    def start(self):
        from aiida.orm.node import Node

        # Testing parameters
        p = self.get_parameters()

        # Testing calculations
        self.attach_calculation(self.generate_calc())
        self.attach_calculation(self.generate_calc())

        # Testing report
        self.append_to_report("Starting workflow with params: {0}".format(p))

        # Testing attachments
        n = Node().store()
        attrs = {"a": [1, 2, 3], "n": n}
        self.add_attributes(attrs)

        # Test process
        self.next(self.second_step)

    @Workflow.step
    def second_step(self):
        # Test retrieval
        calcs = self.get_step_calculations(self.start)
        self.append_to_report("Retrieved calculation 0 (uuid): {0}".format(calcs[0].uuid))

        # Testing report
        a = self.get_attributes()
        self.append_to_report("Execution second_step with attachments: {0}".format(a))

        # Test results
        self.add_result("scf_converged", calcs[0])

        self.next(self.exit)


class SubWorkflowDemo(Workflow):
    def __init__(self, **kwargs):
        super(SubWorkflowDemo, self).__init__(**kwargs)

    def generate_calc(self):
        from aiida.orm import Code, Computer, CalculationFactory
        from aiida.common.datastructures import calc_states

        CustomCalc = CalculationFactory('simpleplugins.templatereplacer')

        computer = Computer.get("localhost")

        calc = CustomCalc(computer=computer, withmpi=True)
        calc.set_resources({"num_machines": 1, "num_mpiprocs_per_machine": 1})
        calc.store()
        calc._set_state(calc_states.FINISHED)
        calc._set_process_state(plumpy.ProcessState.FINISHED)

        return calc

    @Workflow.step
    def start(self):
        self.attach_calculation(self.generate_calc())

        params = {}
        params['nmachine'] = 2

        # Testing subworkflow with input parameters
        w = WorkflowDemo(params=params)
        w.start()
        self.attach_workflow(w)
        self.append_to_report("Workflow attached: {0}".format(w.uuid))

        params['nmachine'] = 4
        w = WorkflowDemo(params=params)
        w.start()
        self.attach_workflow(w)
        self.append_to_report("Workflow attached: {0}".format(w.uuid))

        self.next(self.second)

    @Workflow.step
    def second(self):
        s_wfs = self.get_step(self.start).get_sub_workflows()

        for s_wf in s_wfs:
            self.append_to_report("Workflow {0} has results {1}".format(s_wf.uuid, s_wf.get_result("scf_converged")))

        self.next(self.exit)


class BranchWorkflowDemo(Workflow):
    def __init__(self, **kwargs):

        super(BranchWorkflowDemo, self).__init__(**kwargs)

    def generate_calc(self):

        from aiida.orm import Code, Computer, CalculationFactory
        from aiida.common.datastructures import calc_states

        CustomCalc = CalculationFactory('simpleplugins.templatereplacer')

        computer = Computer.get("localhost")

        calc = CustomCalc(computer=computer, withmpi=True)
        calc.set_resources({"num_machines": 1, "num_mpiprocs_per_machine": 1})
        calc.store()
        calc._set_state(calc_states.FINISHED)
        calc._set_process_state(plumpy.ProcessState.FINISHED)

        return calc

    def start(self):

        self.branch_a_one()
        self.branch_b_one()

    @Workflow.step
    def branch_a_one(self):

        self.append_to_report("branch_a_one launched")

        self.attach_calculation(self.generate_calc())

        self.next(self.branch_a_two)

    @Workflow.step
    def branch_a_two(self):

        self.append_to_report("branch_a_two launched")

        self.attach_calculation(self.generate_calc())

        self.next(self.recollect)

        # @Workflow.step

    #     def branch_a_three(self):
    #
    #         self.append_to_report("branch_a_two launched")
    #
    #         self.attach_calculation(self.generate_calc())
    #
    #         self.next(self.recollect)

    @Workflow.step
    def branch_b_one(self):

        self.append_to_report("branch_b_one launched")

        self.attach_calculation(self.generate_calc())

        self.next(self.recollect)

    @Workflow.step
    def recollect(self):

        self.append_to_report("recollect launched")

        if (self.get_step(self.branch_b_one).has_finished() and
                self.get_step(self.branch_a_two).has_finished()):

            self.append_to_report("All the steps have been done")
            self.next(self.finalize)
        else:

            self.append_to_report("Some step are still running, waiting to recollect")
            self.sleep()

    @Workflow.step
    def finalize(self):
        print "Here in finalize"
        self.append_to_report("Nothing else to do")

        self.next(self.exit)


class LoopBranchWorkflowDemo(Workflow):
    def __init__(self, **kwargs):

        super(LoopBranchWorkflowDemo, self).__init__(**kwargs)

    def generate_calc(self):

        from aiida.orm import Code, Computer, CalculationFactory
        from aiida.common.datastructures import calc_states

        CustomCalc = CalculationFactory('simpleplugins.templatereplacer')

        computer = Computer.get("localhost")

        calc = CustomCalc(computer=computer, withmpi=True)
        calc.set_resources({"num_machines": 1, "num_mpiprocs_per_machine": 1})
        calc.store()
        calc._set_state(calc_states.FINISHED)
        calc._set_process_state(plumpy.ProcessState.FINISHED)

        return calc

    @Workflow.step
    def start(self):

        self.attach_calculation(self.generate_calc())

        self.next(self.convergence)

    @Workflow.step
    def convergence(self):

        calcs_init = self.get_step_calculations(self.start)
        calcs_convergence = self.get_step_calculations(self.convergence)
        tot_calcs_convergence = len(calcs_convergence)

        if calcs_convergence == None or tot_calcs_convergence < 5:
            self.append_to_report("Not enough calculations (only {0}), looping".format(tot_calcs_convergence))
            self.attach_calculation(self.generate_calc())
            self.next(self.convergence)

        else:

            self.append_to_report("Enough calculations run, going to the next step")
            self.next(self.final)

    @Workflow.step
    def final(self):

        calcs_convergence = self.get_step_calculations(self.convergence)

        self.append_to_report("Third ran and retrieved calculation:")
        for c in calcs_convergence:
            self.append_to_report("Calculation {0}".format(c.uuid))

        self.add_result("scf_converged", calcs_convergence[len(calcs_convergence) - 1])

        self.next(self.exit)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
import abc

from aiida.transport import TransportFactory
from aiida.common.exceptions import (ConfigurationError, MissingPluginError)

__all__ = ['AuthInfo', 'AuthInfoCollection']


class AuthInfoCollection(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def create(self, computer, user):
        """
        Create a AuthInfo given a computer and a user

        :param computer: A Computer or DbComputer instance
        :param user: A User or DbUser instance
        :return: a AuthInfo object associated to the given computer and User
        """
        pass

    @abc.abstractmethod
    def get(self, computer, user):
        """
        Return a AuthInfo given a computer and a user

        :param computer: A Computer or DbComputer instance
        :param user: A User or DbUser instance
        :return: a AuthInfo object associated to the given computer and User, if any
        :raise NotExistent: if the user is not configured to use computer
        :raise sqlalchemy.orm.exc.MultipleResultsFound: if the user is configured
             more than once to use the computer! Should never happen
        """
        pass


class AuthInfo(object):
    """
    Base class to map a DbAuthInfo, that contains computer configuration
    specific to a given user (authorization info and other metadata, like
    how often to check on a given computer etc.)
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, backend):
        self._backend = backend

    @property
    def backend(self):
        return self._backend

    def pk(self):
        """
        Return the principal key in the DB.
        """
        return self.id

    @abc.abstractproperty
    def id(self):
        """
        Return the ID in the DB.
        """
        pass

    @abc.abstractproperty
    def enabled(self):
        """
        Is the computer enabled for this user?

        :return: Boolean
        """
        pass

    @enabled.setter
    def enabled(self, value):
        """
        Set the enabled state for the computer

        :return: Boolean
        """
        pass

    @abc.abstractproperty
    def computer(self):
        pass

    @abc.abstractproperty
    def user(self):
        pass

    @abc.abstractproperty
    def is_stored(self):
        """
        Is it already stored or not?

        :return: Boolean
        """
        pass

    @abc.abstractmethod
    def get_auth_params(self):
        """
        Get the dictionary of auth_params

        :return: a dictionary
        """
        pass

    @abc.abstractmethod
    def set_auth_params(self, auth_params):
        """
        Set the dictionary of auth_params

        :param auth_params: a dictionary with the new auth_params
        """
        pass

    @abc.abstractmethod
    def get_metadata(self):
        """
        Get the metadata dictionary

        :return: a dictionary
        """
        pass

    @abc.abstractmethod
    def set_metadata(self, metadata):
        """
        Replace the metadata dictionary in the DB with the provided dictionary
        """
        pass

    def get_workdir(self):
        """
        Get the workdir; defaults to the value of the corresponding computer, if not explicitly set

        :return: a string
        """
        metadata = self.get_metadata()

        try:
            return metadata['workdir']
        except KeyError:
            return self.computer.get_workdir()

    def __str__(self):
        if self.enabled:
            return "AuthInfo for {} on {}".format(self.user.email, self.computer.name)
        else:
            return "AuthInfo for {} on {} [DISABLED]".format(self.user.email, self.computer.name)

    def get_transport(self):
        """
        Return a configured transport to connect to the computer.
        """
        computer = self.computer
        try:
            ThisTransport = TransportFactory(computer.get_transport_type())
        except MissingPluginError as e:
            raise ConfigurationError('No transport found for {} [type {}], message: {}'.format(
                computer.hostname, computer.get_transport_type(), e.message))

        params = dict(computer.get_transport_params().items() +
                      self.get_auth_params().items())
        return ThisTransport(machine=computer.hostname, **params)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Add the process_type column to DbNode

Revision ID: 6c629c886f84
Revises: 0aebbeab274d
Create Date: 2018-03-15 13:23:12.941148

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '6c629c886f84'
down_revision = '0aebbeab274d'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('db_dbnode',
        sa.Column('process_type', sa.VARCHAR(length=255), autoincrement=False, nullable=True),
    )
    op.create_index('ix_db_dbnode_process_type', 'db_dbnode', ['process_type'])


def downgrade():
    op.drop_column('db_dbnode', 'process_type')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from collections import Mapping
from aiida.orm.data import Data
from aiida.orm import load_node


class FrozenDict(Data, Mapping):
    """
    An immutable dictionary containing only Data nodes as values.

    .. note::
        All values must be stored before being passed to constructor.
    """

    def __init__(self, **kwargs):
        self._cache = {}
        self._initialised = False
        super(FrozenDict, self).__init__(**kwargs)
        self._initialised = True

    def set_dict(self, dict):
        assert not self._initialised

        for value in dict.itervalues():
            assert isinstance(value, Data)
            assert value.is_stored

        for k, v in dict.iteritems():
            self._set_attr(k, v.pk)

    def __getitem__(self, key):
        return self._get(key)

    def __iter__(self):
        return self.get_attrs().iterkeys()

    def __len__(self):
        return len(self.get_attrs())

    def _get(self, key):
        val = self._cache.get(key, None)
        if val is None:
            val = load_node(self.get_attr(key))
            self._cache[key] = val
        return val


# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from aiida.common.exceptions import InputValidationError, ValidationError, \
    InvalidOperation
from aiida.restapi.common.exceptions import RestValidationError
from aiida.restapi.translator.base import BaseTranslator


class NodeTranslator(BaseTranslator):
    """
    Translator relative to resource 'nodes' and aiida class Node
    """

    # A label associated to the present class (coincides with the resource name)
    __label__ = "nodes"
    # The AiiDA class one-to-one associated to the present class
    from aiida.orm.node import Node
    _aiida_class = Node
    # The string name of the AiiDA class
    _aiida_type = "node.Node"
    # The string associated to the AiiDA class in the query builder lexicon
    _qb_type = _aiida_type + '.'

    # If True (False) the corresponding AiiDA class has (no) uuid property
    _has_uuid = True

    _result_type = __label__

    _content_type = None

    _alist = None
    _nalist = None
    _elist = None
    _nelist = None
    _downloadformat = None
    _visformat = None
    _filename = None
    _rtype = None


    def __init__(self, Class=None, **kwargs):
        """
        Initialise the parameters.
        Create the basic query_help
        """

        # Assume default class is this class (cannot be done in the
        # definition as it requires self)
        if Class is None:
            Class = self.__class__

        # basic initialization
        super(NodeTranslator, self).__init__(Class=Class, **kwargs)

        self._default_projections = [
            "id",
            "label",
            "type",
            "ctime",
            "mtime",
            "uuid",
            "user_id",
            "user_email",
            "attributes",
            "extras"
        ]

        ## node schema
        # All the values from column_order must present in additional info dict
        # Note: final schema will contain details for only the fields present in column order
        self._schema_projections = {
            "column_order": [
                "id",
                "label",
                "type",
                "ctime",
                "mtime",
                "uuid",
                "user_id",
                "user_email",
                "attributes",
                "extras"
            ],
            "additional_info": {
                "id": {"is_display": True},
                "label": {"is_display": False},
                "type": {"is_display": True},
                "ctime": {"is_display": True},
                "mtime": {"is_display": True},
                "uuid": {"is_display": False},
                "user_id": {"is_display": False},
                "user_email": {"is_display": True},
                "attributes": {"is_display": False},
                "extras": {"is_display": False}
            }
        }

        # Inspect the subclasses of NodeTranslator, to avoid hard-coding
        # (should resemble the following tree)
        """
                                          /- KpointsTranslator
                                         /
                         /- DataTranslator  -- SructureTranslator
                        /                \
                                          \- BandsTranslator
        NodeTranslator  -- CodeTranslator
                        \
                         \- CalculationTranslator
        """

        self._subclasses = self._get_subclasses()

    def set_query_type(self, query_type, alist=None, nalist=None, elist=None,
                       nelist=None, downloadformat=None, visformat=None,
                       filename=None, rtype=None):
        """
        sets one of the mutually exclusive values for self._result_type and
        self._content_type.

        :param query_type:(string) the value assigned to either variable.
        """

        if query_type == "default":
            pass
        elif query_type == "inputs":
            self._result_type = 'input_of'
        elif query_type == "outputs":
            self._result_type = "output_of"
        elif query_type == "attributes":
            self._content_type = "attributes"
            self._alist = alist
            self._nalist = nalist
        elif query_type == "extras":
            self._content_type = "extras"
            self._elist = elist
            self._nelist = nelist
        elif query_type == 'visualization':
            self._content_type = 'visualization'
            self._visformat = visformat
        elif query_type == 'download':
            self._content_type = 'download'
            self._downloadformat = downloadformat
        elif query_type == "retrieved_inputs":
            self._content_type = 'retrieved_inputs'
            self._filename = filename
            self._rtype = rtype
        elif query_type == "retrieved_outputs":
            self._content_type = 'retrieved_outputs'
            self._filename = filename
            self._rtype = rtype
        else:
            raise InputValidationError("invalid result/content value: {"
                                       "}".format(query_type))

        ## Add input/output relation to the query help
        if self._result_type is not self.__label__:
            self._query_help["path"].append(
                {
                    "type": "node.Node.",
                    "label": self._result_type,
                    self._result_type: self.__label__
                })

    def set_query(self, filters=None, orders=None, projections=None,
                  query_type=None, id=None, alist=None, nalist=None,
                  elist=None, nelist=None, downloadformat=None, visformat=None,
                  filename=None, rtype=None):
        """
        Adds filters, default projections, order specs to the query_help,
        and initializes the qb object

        :param filters: dictionary with the filters
        :param orders: dictionary with the order for each tag
        :param projections: dictionary with the projection. It is discarded
            if query_type=='attributes'/'extras'
        :param query_type: (string) specify the result or the content ("attr")
        :param id: (integer) id of a specific node
        """

        ## Check the compatibility of query_type and id
        if query_type is not "default" and id is None:
            raise ValidationError("non default result/content can only be "
                                  "applied to a specific node (specify an id)")

        ## Set the type of query
        self.set_query_type(query_type, alist=alist, nalist=nalist,
                            elist=elist, nelist=nelist, downloadformat=downloadformat,
                            visformat=visformat, filename=filename, rtype=rtype)

        ## Define projections
        if self._content_type is not None:
            # Use '*' so that the object itself will be returned.
            # In get_results() we access attributes/extras by
            # calling the get_attrs()/get_extras().
            projections = ['*']
        else:
            pass  # i.e. use the input parameter projection

        # TODO this actually works, but the logic is a little bit obscure.
        # Make it clearer
        if self._result_type is not self.__label__:
            projections = self._default_projections

        super(NodeTranslator, self).set_query(filters=filters,
                                              orders=orders,
                                              projections=projections,
                                              id=id)

    def _get_content(self):
        """
        Used by get_results() in case of endpoint include "content" option
        :return: data: a dictionary containing the results obtained by
        running the query
        """
        if not self._is_qb_initialized:
            raise InvalidOperation("query builder object has not been "
                                   "initialized.")

        ## Count the total number of rows returned by the query (if not
        # already done)
        if self._total_count is None:
            self.count()

        # If count==0 return
        if self._total_count == 0:
            return {}

        # otherwise ...
        n = self.qb.first()[0]

        # content/attributes
        if self._content_type == "attributes":
            # Get all attrs if nalist and alist are both None
            if self._alist is None and self._nalist is None:
                data = {self._content_type: n.get_attrs()}
            # Get all attrs except those contained in nalist
            elif self._alist is None and self._nalist is not None:
                attrs = {}
                for key in n.get_attrs().keys():
                    if key not in self._nalist:
                        attrs[key] = n.get_attr(key)
                data = {self._content_type: attrs}
            # Get all attrs contained in alist
            elif self._alist is not None and self._nalist is None:
                attrs = {}
                for key in n.get_attrs().keys():
                    if key in self._alist:
                        attrs[key] = n.get_attr(key)
                data = {self._content_type: attrs}
            else:
                raise RestValidationError("you cannot specify both alist "
                                          "and nalist")
        # content/extras
        elif self._content_type == "extras":

            # Get all extras if nelist and elist are both None
            if self._elist is None and self._nelist is None:
                data = {self._content_type: n.get_extras()}

            # Get all extras except those contained in nelist
            elif self._elist is None and self._nelist is not None:
                extras = {}
                for key in n.get_extras().keys():
                    if key not in self._nelist:
                        extras[key] = n.get_extra(key)
                data = {self._content_type: extras}

            # Get all extras contained in elist
            elif self._elist is not None and self._nelist is None:
                extras = {}
                for key in n.get_extras().keys():
                    if key in self._elist:
                        extras[key] = n.get_extra(key)
                data = {self._content_type: extras}

            else:
                raise RestValidationError("you cannot specify both elist "
                                          "and nelist")

        # Data needed for visualization appropriately serialized (this
        # actually works only for data derived classes)
        # TODO refactor the code so to have this option only in data and
        # derived classes
        elif self._content_type == 'visualization':
            # In this we do not return a dictionary but just an object and
            # the dictionary format is set by get_visualization_data
            data = {self._content_type: self.get_visualization_data(n, self._visformat)}

        elif self._content_type == 'download':
            # In this we do not return a dictionary but download the file in
            # specified format if available
            data = {self._content_type: self.get_downloadable_data(n, self._downloadformat)}

        elif self._content_type == 'retrieved_inputs':
            # This type is only available for calc nodes. In case of job calc it
            # returns calc inputs prepared to submit calc on the cluster else []
            data = {self._content_type: self.get_retrieved_inputs(n, self._filename, self._rtype)}

        elif self._content_type == 'retrieved_outputs':
            # This type is only available for calc nodes. In case of job calc it
            # returns calc outputs retrieved from the cluster else []
            data = {self._content_type: self.get_retrieved_outputs(n, self._filename, self._rtype)}

        else:
            raise ValidationError("invalid content type")

        return data

    def _get_subclasses(self, parent=None, parent_class=None, recursive=True):
        """
        Import all submodules of the package containing the present class.
        
        Includes subpackages recursively, if specified.

        :param parent: package/class. 
            If package looks for the classes in submodules.
            If class, first lookss for the package where it is contained
        :param parent_class: class of which to look for subclasses
        :param recursive: True/False (go recursively into submodules)
        """

        import pkgutil
        import imp
        import inspect
        import os

        # If no parent class is specified set it to self.__class
        parent = self.__class__ if parent is None else parent

        # Suppose parent is class
        if inspect.isclass(parent):

            # Set the package where the class is contained
            classfile = inspect.getfile(parent)
            package_path = os.path.dirname(classfile)

            # If parent class is not specified, assume it is the parent
            if parent_class is None:
                parent_class = parent

        # Suppose parent is a package (directory containing __init__.py).
        # Check if it contains attribute __path__
        elif inspect.ismodule(parent) and hasattr(parent, '__path__'):

            # Assume path is one element list
            package_path = parent.__path__[0]

            # if parent is a package, parent_class cannot be None
            if parent_class is None:
                raise TypeError('argument parent_class cannot be None')

                # Recursively check the subpackages
        results = {}

        for loader, name, is_pkg in pkgutil.walk_packages([package_path]):
            # N.B. pkgutil.walk_package requires a LIST of paths.

            full_path_base = os.path.join(package_path, name)

            if is_pkg:
                app_module = imp.load_package(full_path_base, full_path_base)
            else:
                full_path = full_path_base + '.py'
                # I could use load_module but it takes lots of arguments,
                # then I use load_source
                app_module = imp.load_source("rst"+name, full_path)

            # Go through the content of the module
            if not is_pkg:
                for name, obj in inspect.getmembers(app_module):
                    if inspect.isclass(obj) and issubclass(obj, parent_class):
                        results[name] = obj
            # Look in submodules
            elif is_pkg and recursive:
                results.update(self._get_subclasses(parent=app_module,
                                                    parent_class=parent_class))

        return results

    def get_visualization_data(self, node, format=None):
        """
        Generic function to get the data required to visualize the node with
        a specific plugin.
        Actual definition is in child classes as the content to be be
        returned and its format depends on the visualization plugin specific
        to the resource

        :param node: node object that has to be visualized
        :returns: data selected and serializaed for visualization
        """

        """
        If this method is called by Node resource it will look for the type
        of object and invoke the correct method in the lowest-compatibel
        subclass
        """

        # Look for the translator associated to the class of which this node
        # is instance
        tclass = type(node)

        for subclass in self._subclasses.values():
            if subclass._aiida_type.split('.')[-1] == tclass.__name__:
                lowtrans = subclass

        visualization_data = lowtrans.get_visualization_data(node, format=format)

        return visualization_data

    def get_downloadable_data(self, node, format=None):
        """
        Generic function to download file in specified format.
        Actual definition is in child classes as the content to be
        returned and its format depends on the visualization plugin specific
        to the resource

        :param node: node object
        :param format: file extension format
        :returns: data in selected format to download
        """

        """
        If this method is called by Node resource it will look for the type
        of object and invoke the correct method in the lowest-compatibel
        subclass
        """

        # Look for the translator associated to the class of which this node
        # is instance
        tclass = type(node)

        for subclass in self._subclasses.values():
            if subclass._aiida_type.split('.')[-1] == tclass.__name__:
                lowtrans = subclass

        downloadable_data = lowtrans.get_downloadable_data(node, format=format)

        return downloadable_data

    def get_retrieved_inputs(self, node, filename=None, rtype=None):
        """
        Generic function to return output of calc inputls verdi command.
        Actual definition is in child classes as the content to be
        returned and its format depends on the visualization plugin specific
        to the resource

        :param node: node object
        :returns: list of calc inputls command
        """

        if node.type.startswith("calculation"):
            from aiida.restapi.translator.calculation import CalculationTranslator
            return CalculationTranslator.get_retrieved_inputs(node, filename=filename, rtype=rtype)
        return []

    def get_retrieved_outputs(self, node, filename=None, rtype=None):
        """
        Generic function to return output of calc outputls verdi command.
        Actual definition is in child classes as the content to be
        returned and its format depends on the visualization plugin specific
        to the resource

        :param node: node object
        :returns: list of calc outputls command
        """

        if node.type.startswith("calculation"):
            from aiida.restapi.translator.calculation import CalculationTranslator
            return CalculationTranslator.get_retrieved_outputs(node, filename=filename, rtype=rtype)
        return []

    @staticmethod
    def get_file_content(node, file_name):
        """
        It reads the file from directory and returns its content.

        Instead of using "send_from_directory" from flask, this method is written
        because in next aiida releases the file can be stored locally or in object storage.

        :param node: aiida folderData node which contains file
        :param file_name: name of the file to return its contents
        :return:
        """
        import os
        file_parts = file_name.split(os.sep)

        if len(file_parts) > 1:
            file_name = file_parts.pop()
            for folder in file_parts:
                node = node.get_subfolder(folder)

        with node.open(file_name) as f:
            return f.read()

    def get_results(self):
        """
        Returns either a list of nodes or details of single node from database

        :return: either a list of nodes or the details of single node
            from the database
        """
        if self._content_type is not None:
            return self._get_content()
        else:
            return super(NodeTranslator, self).get_results()

    def get_statistics(self, user_pk=None):
        """Return statistics for a given node"""

        from aiida.backends.utils import QueryFactory
        qmanager = QueryFactory()()
        return qmanager.get_creation_statistics(user_pk=user_pk)


    def get_io_tree(self, uuid_pattern):
        from aiida.orm.querybuilder import QueryBuilder
        from aiida.orm.node import Node

        def get_node_shape(ntype):
            type = ntype.split(".")[0]

            # default and data node shape
            shape = "dot"

            if type == "calculation":
                shape = "square"
            elif type == "code":
                shape = "triangle"

            return shape

        # Check whether uuid_pattern identifies a unique node
        self._check_id_validity(uuid_pattern)

        qb = QueryBuilder()
        qb.append(Node, tag="main", project=["*"],
                  filters=self._id_filter)

        nodes = []
        edges = []
        nodeCount = 0

        if qb.count() > 0:
            mainNode = qb.first()[0]
            pk = mainNode.pk
            uuid = mainNode.uuid
            nodetype = mainNode.type
            display_type = nodetype.split('.')[-2]
            description = mainNode.get_desc()
            if description == '':
                description = mainNode.type.split('.')[-2]

            nodes.append({
                "id": nodeCount,
                "nodeid": pk,
                "nodeuuid": uuid,
                "nodetype": nodetype,
                "displaytype": display_type,
                "group": "mainNode",
                "description": description,
                "shape": get_node_shape(nodetype)
            })
        nodeCount += 1

        # get all inputs
        qb = QueryBuilder()
        qb.append(Node, tag="main", project=['*'],
                  filters=self._id_filter)
        qb.append(Node, tag="in", project=['*'], edge_project=['label'],
                  input_of='main')

        if qb.count() > 0:
            for input in qb.iterdict():
                node = input['in']['*']
                linktype = input['main--in']['label']
                pk = node.pk
                uuid = node.uuid
                nodetype = node.type
                display_type = nodetype.split('.')[-2]
                description = node.get_desc()
                if description == '':
                    description = node.type.split('.')[-2]

                nodes.append({
                    "id": nodeCount,
                    "nodeid": pk,
                    "nodeuuid": uuid,
                    "nodetype": nodetype,
                    "displaytype": display_type,
                    "group": "inputs",
                    "description": description,
                    "linktype": linktype,
                    "shape": get_node_shape(nodetype)
                })
                edges.append({
                    "from": nodeCount,
                    "to": 0,
                    "arrows": "to",
                    "color": {"inherit": 'from'},
                    "linktype": linktype,
                })
                nodeCount += 1

        # get all outputs
        qb = QueryBuilder()
        qb.append(Node, tag="main", project=['*'],
                  filters=self._id_filter)
        qb.append(Node, tag="out", project=['*'], edge_project=['label'],
                  output_of='main')
        if qb.count() > 0:
            for output in qb.iterdict():
                node = output['out']['*']
                linktype = output['main--out']['label']
                pk = node.pk
                uuid = node.uuid
                nodetype = node.type
                display_type = nodetype.split('.')[-2]
                description = node.get_desc()
                if description == '':
                    description = node.type.split('.')[-2]

                nodes.append({
                    "id": nodeCount,
                    "nodeid": pk,
                    "nodeuuid": uuid,
                    "nodetype": nodetype,
                    "displaytype": display_type,
                    "group": "outputs",
                    "description": description,
                    "linktype": linktype,
                    "shape": get_node_shape(nodetype)
                })
                edges.append({
                    "from": 0,
                    "to": nodeCount,
                    "arrows": "to",
                    "color": {"inherit": 'to'},
                    "linktype": linktype
                })
                nodeCount += 1

        return {"nodes": nodes, "edges": edges}



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

install_requires = [
    'reentry==1.2.0',
    'python-dateutil==2.7.2',
    'python-mimeparse==1.6.0',
    'django==1.7.11',  # Upgrade to Django 1.9 does prevent AiiDA functioning
    'django-extensions==1.5.0',
    'tzlocal==1.5.1',
    'pytz==2018.4',
    'PyYAML==3.12',
    'six==1.11.0',
    'future==0.16.0',
    'anyjson==0.3.3',
    'psutil==5.4.5',
    'meld3==1.0.2',
    'numpy==1.14.3',
    'SQLAlchemy==1.0.19',  # Upgrade to SQLalchemy 1.1.5 does break tests, see #465
    'SQLAlchemy-Utils==0.33.0',
    'alembic==0.9.9',
    'ujson==1.35',
    'enum34==1.1.6',
    'aldjemy==0.8.0',
    'passlib==1.7.1',
    'validate-email==1.3',
    'click==6.7',
    'click-plugins==1.0.3',
    'click-spinner==0.1.7',
    'tabulate==0.8.2',
    'ete3==3.1.1',
    'uritools==2.1.0',
    'psycopg2-binary==2.7.4',
    'paramiko==2.4.1',
    'ecdsa==0.13',
    'pycrypto==2.6.1',
    'pika==0.11.2',
    'ipython<6.0',  # Version of ipython non enforced, because some still prefer version 4 rather than the latest
    'plumpy==0.10.3',
    'circus==0.14.0',
    'tornado==4.5.3',  # As of 2018/03/06 Tornado released v5.0 which breaks circus 0.14.0
]

extras_require = {
    ':python_version < "3.3"': ['mock'],
    ':python_version < "3"': ['chainmap', 'pathlib2', 'singledispatch >= 3.4.0.3'],
    # Requirements for ssh transport with authentification through Kerberos token
    # N. B.: you need to install first libffi and MIT kerberos GSSAPI including header files.
    # E.g. for Ubuntu 14.04: sudo apt-get install libffi-dev libkrb5-dev
    'ssh_kerberos': [
        'pyasn1==0.4.2',
        'python-gssapi==0.6.4',
    ],
    # Requirements for RESTful API
    'rest': [
        'Flask==1.0.2',
        'Flask-RESTful==0.3.6',
        'Flask-Cors==3.0.4',
        'pyparsing==2.2.0',
        'Pattern==2.6',
        'Flask-SQLAlchemy==2.3.2',
        'sqlalchemy-migrate==0.11.0',
        'marshmallow-sqlalchemy==0.13.2',
        'flask-marshmallow==0.9.0',
        'itsdangerous==0.24',
        'Flask-HTTPAuth==3.2.3',
        'Flask-Cache==0.13.1',
        'python-memcached==1.59',
    ],
    # Requirements to building documentation
    'docs': [
        'Sphinx==1.7.4',
        'Pygments==2.2.0',
        'docutils==0.14',
        'Jinja2==2.10',
        'MarkupSafe==1.0',
        'sphinx-rtd-theme==0.3.1',  # Required by readthedocs
    ],
    # Requirements for non-core functionalities that rely on external atomic manipulation/processing software
    'atomic_tools': [
        'spglib==1.10.3.65',
        'pymatgen==2018.4.20',
        'ase==3.12.0',  # Updating breaks tests
        'PyMySQL==0.8.0',  # Required by ICSD tools
        'PyCifRW==4.2.1',  # Updating breaks tests
        'seekpath==1.8.1',
        'qe-tools==1.1.0',
    ],
    # Requirements for jupyter notebook
    'notebook': [
        'jupyter==1.0.0',
    ],
    # Requirements for testing
    'testing': [
        'mock==2.0.0',
        'pgtest==1.1.0',
        'sqlalchemy-diff==0.1.3',
        'coverage==4.5.1',
        'codecov==2.0.15'
    ],
    'dev_precommit': [
        'pre-commit==1.8.2',
        'yapf==0.21.0',
        'prospector==0.12.7',
        'pylint==1.8.4',
        'pep8-naming==0.3.3',
        'toml==0.9.4'
    ],
    'dev_sphinxext': [
        'pytest==3.5.1',
        'pytest-cov==2.5.1',
    ]
}

extras_require['dev_sphinxext'] += extras_require['docs']
extras_require['testing'] += extras_require['rest'] + extras_require['atomic_tools'] + extras_require['dev_sphinxext']
extras_require['all'] = [item for sublist in extras_require.values() for item in sublist]

# There are a number of optional dependencies that are not
# listed even as optional dependencies as they are quite
# cumbersome to install and there is a risk that a user, wanting
# to install all dependencies (including optional ones)
# does not manage and thinks it's an AiiDA problem.
#
# These include:
#  - mayavi>=4.5.0
#    plotting package, requires to have the vtk code installed first;
#    moreover requires to have numpy installed before, but it is not in
#    the requirements (and there is no easy way on our side to fix a specific
#    installation order of dependencies)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from aiida.common.extendeddicts import FixedFieldsAttributeDict
from aiida.plugins.factory import BaseFactory


def TransportFactory(entry_point):
    """
    Return the Transport plugin class for a given entry point

    :param entry_point: the entry point name of the Transport plugin
    """
    return BaseFactory('aiida.transports', entry_point)


class FileAttribute(FixedFieldsAttributeDict):
    """
    A class, resembling a dictionary, to describe the attributes of a file,
    that is returned by get_attribute().
    Possible keys: st_size, st_uid, st_gid, st_mode, st_atime, st_mtime
    """
    _valid_fields = (
        'st_size',
        'st_uid',
        'st_gid',
        'st_mode',
        'st_atime',
        'st_mtime',
    )



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
import os
import uuid
import json
import unittest
try:
    from unittest import mock
except ImportError:
    import mock

from ._utils import check_and_migrate_config
from ._migrations import _MIGRATION_LOOKUP


def load_config_sample(filename):
    currdir = os.path.dirname(os.path.abspath(__file__))
    filepath = os.path.join(currdir, 'test_samples', filename)
    with open(filepath, 'r') as f:
        return json.load(f)


class TestConfigMigration(unittest.TestCase):

    def test_check_and_migrate(self):
        """
        Test the full config migration.
        """
        initial_config = load_config_sample('input/0.json')
        with mock.patch.object(uuid, 'uuid4', return_value=uuid.UUID(hex='0'*32)):
            final_config = check_and_migrate_config(initial_config, store=False)
        reference_config = load_config_sample('reference/final.json')
        self.assertEqual(final_config, reference_config)

    def test_0_1_migration(self):
        """
        Test the step between config versions 0 and 1.
        """
        initial_config = load_config_sample('input/0.json')
        reference_config = load_config_sample('reference/1.json')
        final_config = _MIGRATION_LOOKUP[0].apply(initial_config)
        self.assertEqual(final_config, reference_config)

    def test_1_2_migration(self):
        """
        Test the step between config versions 1 and 2.
        """
        self.maxDiff = None
        initial_config = load_config_sample('input/1.json')
        reference_config = load_config_sample('reference/2.json')
        with mock.patch.object(uuid, 'uuid4', return_value=uuid.UUID(hex='0'*32)):
            final_config = _MIGRATION_LOOKUP[1].apply(initial_config)
        self.assertEqual(final_config, reference_config)

    def test_2_3_migration(self):
        """
        Test the step between config versions 2 and 3.
        """
        initial_config = load_config_sample('input/2.json')
        reference_config = load_config_sample('reference/3.json')
        final_config = _MIGRATION_LOOKUP[2].apply(initial_config)
        self.assertEqual(final_config, reference_config)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

import copy
import unittest

import os
from alembic import command
from alembic.config import Config

from aiida.backends import sqlalchemy as sa
from aiida.backends.sqlalchemy import utils
from aiida.backends.sqlalchemy.models.base import Base
from aiida.backends.sqlalchemy.utils import (get_migration_head,
                                             get_db_schema_version)
from aiida.backends.testbase import AiidaTestCase
from aiida.common.setup import set_property, get_property

from aiida.backends.sqlalchemy.tests.utils import new_database


alembic_root = os.path.join(os.path.dirname(__file__), 'migrations', 'alembic')


TEST_ALEMBIC_REL_PATH = 'migration_test'


class TestMigrationApplicationSQLA(AiidaTestCase):
    """
    This class contains tests for the migration mechanism of SQLAlchemy called
    alembic. It checks if the migrations can be applied and removed correctly.
    """
    # The path to the folder that contains the migration configuration (the
    # actual configuration - not the testing)
    migr_method_dir_path = None
    # The path of the migration configuration (the actual configuration - not
    # the testing)
    alembic_dpath = None

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        super(TestMigrationApplicationSQLA, cls).setUpClass(*args, **kwargs)
        cls.migr_method_dir_path = os.path.dirname(
            os.path.realpath(utils.__file__))

    def setUp(self):
        self.migrate_db_with_non_testing_migrations("base")

    def tearDown(self):
        self.migrate_db_with_non_testing_migrations("head")

    def migrate_db_with_non_testing_migrations(self, destination):
        if destination not in ["head", "base"]:
            raise TypeError("Only head & base are accepted as destination "
                            "values.")
        # Set the alembic script directory location
        self.alembic_dpath = os.path.join(self.migr_method_dir_path,
                                     utils.ALEMBIC_REL_PATH)
        alembic_cfg = Config()
        alembic_cfg.set_main_option('script_location', self.alembic_dpath)

        # Undo all previous real migration of the database
        with sa.engine.begin() as connection:
            alembic_cfg.attributes['connection'] = connection
            if destination == "head":
                command.upgrade(alembic_cfg, "head")
            else:
                command.downgrade(alembic_cfg, "base")

    def test_migrations_forward_backward(self):
        """
        This is a very broad test that checks that the migration mechanism
        works. More specifically, it checks that::

            - Alembic database migrations to specific versions work (upgrade &
              downgrade)

            - The methods that are checking the database schema version and perform
              the migration procedure to the last version work correctly.

        """
        from aiida.backends.sqlalchemy.tests.migration_test import versions
        from aiida.backends.sqlalchemy.utils import check_schema_version

        try:
            # Constructing the versions directory
            versions_dpath = os.path.join(
                os.path.dirname(versions.__file__))

            # Setting dynamically the the path to the alembic configuration
            # (this is where the env.py file can be found)
            alembic_cfg = Config()
            alembic_cfg.set_main_option('script_location', self.alembic_dpath)
            # Setting dynamically the versions directory. These are the
            # migration scripts to pass from one version to the other. The
            # default ones are overridden with test-specific migrations.
            alembic_cfg.set_main_option('version_locations', versions_dpath)

            # Using the connection initialized by the tests
            with sa.engine.begin() as connection:
                alembic_cfg.attributes['connection'] = connection

                self.assertIsNone(get_db_schema_version(alembic_cfg),
                                  "The initial database version should be "
                                  "None (no version) since the test setUp "
                                  "method should undo all migrations")
            # Migrate the database to the latest version
            check_schema_version(force_migration=True, alembic_cfg=alembic_cfg)
            with sa.engine.begin() as connection:
                alembic_cfg.attributes['connection'] = connection
                self.assertEquals(get_db_schema_version(alembic_cfg),
                                  get_migration_head(alembic_cfg),
                                  "The latest database version is not the "
                                  "expected one.")
            with sa.engine.begin() as connection:
                alembic_cfg.attributes['connection'] = connection
                # Migrating the database to the base version
                command.downgrade(alembic_cfg, "base")
                self.assertIsNone(get_db_schema_version(alembic_cfg),
                                  "The database version is not the expected "
                                  "one. It should be None (initial).")

        except Exception as test_ex:
            # If there is an exception, clean the alembic related tables
            from sqlalchemy.engine import reflection

            # Getting the current database table names
            inspector = reflection.Inspector.from_engine(
                sa.get_scoped_session().bind)
            db_table_names = set(inspector.get_table_names())
            # The alembic related database names
            alemb_table_names = set(['account', 'alembic_version'])

            # Get the intersection of the above tables
            tables_to_drop = set.intersection(db_table_names,
                                              alemb_table_names)
            # Delete only the tables that exist
            for table in tables_to_drop:
                from psycopg2 import ProgrammingError
                from sqlalchemy.orm import sessionmaker, scoped_session
                try:
                    with sa.engine.begin() as connection:
                        connection.execute('DROP TABLE {};'.format(table))
                except Exception as db_ex:
                    print("The following error occured during the cleaning of"
                          "the database: {}".format(db_ex.message))
            # Since the database cleaning is over, raise the test
            # exception that was caught
            raise test_ex


class TestMigrationSchemaVsModelsSchema(unittest.TestCase):
    """
    This class checks that the schema that results from a migration is the
    same generated by the models. This is important since migrations are
    frequently written by hand or extended manually and we have to ensure
    that the final result is what is conceived in the SQLA models.
    """
    # The path to the folder that contains the migration configuration (the
    # actual configuration - not the testing)
    migr_method_dir_path = None
    # The path of the migration configuration (the actual configuration - not
    # the testing)
    alembic_dpath = None
    # The alembic configuration needed for the migrations is stored here
    alembic_cfg_left = None

    # The URL of the databases
    db_url_left = None
    db_url_right = None

    def setUp(self):
        # from aiida.backends.sqlalchemy.tests.migration_test import versions
        from sqlalchemydiff.util import get_temporary_uri
        from aiida.backends.sqlalchemy.migrations import versions

        self.migr_method_dir_path = os.path.dirname(
            os.path.realpath(utils.__file__))
        # Set the alembic script directory location
        self.alembic_dpath = os.path.join(self.migr_method_dir_path,
                                     utils.ALEMBIC_REL_PATH)

        # Constructing the versions directory
        versions_dpath = os.path.join(
            os.path.dirname(versions.__file__))

        # Setting dynamically the the path to the alembic configuration
        # (this is where the env.py file can be found)
        self.alembic_cfg_left = Config()
        self.alembic_cfg_left.set_main_option('script_location', self.alembic_dpath)
        # Setting dynamically the versions directory. These are the
        # migration scripts to pass from one version to the other. The
        # default ones are overridden with test-specific migrations.
        self.alembic_cfg_left.set_main_option('version_locations', versions_dpath)

        # The correction URL to the SQLA database of the current
        # AiiDA connection
        curr_db_url = sa.engine.url

        # Create new urls for the two new databases
        self.db_url_left = get_temporary_uri(str(curr_db_url))
        self.db_url_right = get_temporary_uri(str(curr_db_url))

        # Put the correct database url to the database used by alembic
        self.alembic_cfg_left.set_main_option("sqlalchemy.url",
                                              self.db_url_left)

        # Database creation
        new_database(self.db_url_left)
        new_database(self.db_url_right)

    def tearDown(self):
        from sqlalchemydiff.util import destroy_database
        destroy_database(self.db_url_left)
        destroy_database(self.db_url_right)

    def test_model_and_migration_schemas_are_the_same(self):
        """Compare two databases.

        Compares the database obtained with all migrations against the
        one we get out of the models.  It produces a text file with the
        results to help debug differences.
        """
        from sqlalchemy.engine import create_engine
        from sqlalchemydiff import compare

        with create_engine(self.db_url_left).begin() as connection:
            self.alembic_cfg_left.attributes['connection'] = connection
            command.upgrade(self.alembic_cfg_left, "head")

        engine_right = create_engine(self.db_url_right)
        Base.metadata.create_all(engine_right)
        engine_right.dispose()

        result = compare(
            self.db_url_left, self.db_url_right, set(['alembic_version']))

        self.assertTrue(result.is_match,
                        "The migration database doesn't match to the one "
                        "created by the models.\nDifferences: " +
                        result._dump_data(result.errors))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Verdi daemon commands
"""
import os
import sys
import subprocess
import time

import click
from click_spinner import spinner

from aiida.cmdline.baseclass import VerdiCommandWithSubcommands
from aiida.cmdline.commands import verdi, daemon_cmd
from aiida.cmdline.utils import decorators
from aiida.cmdline.utils.common import get_env_with_venv_bin
from aiida.cmdline.utils.daemon import get_daemon_status, print_client_response_status
from aiida.common.profile import get_current_profile_name
from aiida.common.setup import get_profiles_list
from aiida.daemon.client import DaemonClient


class Daemon(VerdiCommandWithSubcommands):
    """
    Manage the AiiDA daemon

    This command allows to interact with the AiiDA daemon.
    Valid subcommands are:

    * start: start the daemon

    * stop: restart the daemon

    * restart: restart the aiida daemon, waiting for it to cleanly exit\
        before restarting it.

    * status: inquire the status of the Daemon.

    * logshow: show the log in a continuous fashion, similar to the 'tail -f' \
        command. Press CTRL+C to exit.
    """

    def __init__(self):
        """
        A dictionary with valid commands and functions to be called:
        start, stop, status and restart.
        """
        super(Daemon, self).__init__()

        self.valid_subcommands = {
            'start': (self.cli, self.complete_none),
            'status': (self.cli, self.complete_none),
            'restart': (self.cli, self.complete_none),
            'stop': (self.cli, self.complete_none),
            'incr': (self.cli, self.complete_none),
            'decr': (self.cli, self.complete_none),
            'logshow': (self.cli, self.complete_none),
            '_start_circus': (self.cli, self.complete_none),
        }

    def cli(self, *args):  # pylint: disable=unused-argument,no-self-use
        verdi.main()


@daemon_cmd.command()
@click.option('--foreground', is_flag=True, help='Run in foreground')
@decorators.with_dbenv()
@decorators.check_circus_zmq_version
def start(foreground):
    """
    Start the daemon
    """
    client = DaemonClient()

    click.echo('Starting the daemon... ', nl=False)

    if foreground:
        command = ['verdi', '-p', client.profile_name, 'daemon', '_start_circus', '--foreground']
    else:
        command = ['verdi', '-p', client.profile_name, 'daemon', '_start_circus']

    try:
        currenv = get_env_with_venv_bin()
        subprocess.check_output(command, env=currenv)
    except subprocess.CalledProcessError as exception:
        click.echo('failed: {}'.format(exception))
        sys.exit(1)

    # We add a small timeout to give the pid-file a chance to be created
    with spinner():
        time.sleep(1)
        response = client.get_status()

    print_client_response_status(response)


@daemon_cmd.command()
@click.option('--all', 'all_profiles', is_flag=True, help='Show all daemons')
def status(all_profiles):
    """
    Print the status of the current daemon or all daemons
    """
    if all_profiles is True:
        profiles = [p for p in get_profiles_list() if not p.startswith('test_')]
    else:
        profiles = [get_current_profile_name()]

    for profile_name in profiles:
        client = DaemonClient(profile_name)
        click.secho('Profile: ', fg='red', bold=True, nl=False)
        click.secho('{}'.format(profile_name), bold=True)
        result = get_daemon_status(client)
        click.echo(result)


@daemon_cmd.command()
@click.argument('number', default=1, type=int)
@decorators.only_if_daemon_pid
def incr(number):
    """
    Add NUMBER [default=1] workers to the running daemon
    """
    client = DaemonClient()
    response = client.increase_workers(number)
    print_client_response_status(response)


@daemon_cmd.command()
@click.argument('number', default=1, type=int)
@decorators.only_if_daemon_pid
def decr(number):
    """
    Remove NUMBER [default=1] workers from the running daemon
    """
    client = DaemonClient()
    response = client.decrease_workers(number)
    print_client_response_status(response)


@daemon_cmd.command()
def logshow():
    """
    Show the log of the daemon, press CTRL+C to quit
    """
    client = DaemonClient()

    try:
        currenv = get_env_with_venv_bin()
        process = subprocess.Popen(['tail', '-f', client.daemon_log_file], env=currenv)
        process.wait()
    except KeyboardInterrupt:
        process.kill()


@daemon_cmd.command()
@click.option('--no-wait', is_flag=True, help='Do not wait for confirmation')
@click.option('--all', 'all_profiles', is_flag=True, help='Stop all daemons')
def stop(no_wait, all_profiles):
    """
    Stop the daemon
    """
    if all_profiles is True:
        profiles = [p for p in get_profiles_list() if not p.startswith('test_')]
    else:
        profiles = [get_current_profile_name()]

    for profile_name in profiles:

        client = DaemonClient(profile_name)

        click.secho('Profile: ', fg='red', bold=True, nl=False)
        click.secho('{}'.format(profile_name), bold=True)

        if not client.is_daemon_running:
            click.echo('Daemon was not running')
            continue

        wait = not no_wait

        if wait:
            click.echo('Waiting for the daemon to shut down... ', nl=False)
        else:
            click.echo('Shutting the daemon down')

        response = client.stop_daemon(wait)

        if wait:
            print_client_response_status(response)


@daemon_cmd.command()
@click.option('--reset', is_flag=True, help='Completely reset the daemon')
@click.option('--no-wait', is_flag=True, help='Do not wait for confirmation')
@click.pass_context
@decorators.with_dbenv()
@decorators.only_if_daemon_pid
def restart(ctx, reset, no_wait):
    """
    Restart the daemon. By default will only reset the workers of the running daemon.
    After the restart the same amount of workers will be running. If the --reset flag
    is passed, however, the full circus daemon will be stopped and restarted with just
    a single worker
    """
    client = DaemonClient()

    wait = not no_wait

    if reset:
        ctx.invoke(stop)
        ctx.invoke(start)
    else:

        if wait:
            click.echo('Restarting the daemon... ', nl=False)
        else:
            click.echo('Restarting the daemon')

        response = client.restart_daemon(wait)

        if wait:
            print_client_response_status(response)


@daemon_cmd.command()
@click.option('--foreground', is_flag=True, help='Run in foreground')
@decorators.with_dbenv()
@decorators.check_circus_zmq_version
def _start_circus(foreground):
    """
    This will actually launch the circus daemon, either daemonized in the background
    or in the foreground, printing all logs to stdout.

    .. note:: this should not be called directly from the commandline!
    """
    from circus import get_arbiter
    from circus import logger as circus_logger
    from circus.circusd import daemonize
    from circus.pidfile import Pidfile
    from circus.util import check_future_exception_and_log, configure_logger

    client = DaemonClient()

    loglevel = client.loglevel
    logoutput = '-'

    if not foreground:
        logoutput = client.circus_log_file

    arbiter_config = {
        'controller': client.get_controller_endpoint(),
        'pubsub_endpoint': client.get_pubsub_endpoint(),
        'stats_endpoint': client.get_stats_endpoint(),
        'logoutput': logoutput,
        'loglevel': loglevel,
        'debug': False,
        'statsd': True,
        'pidfile': client.circus_pid_file,
        'watchers': [{
            'name': client.daemon_name,
            'cmd': client.cmd_string,
            'virtualenv': client.virtualenv,
            'copy_env': True,
            'stdout_stream': {
                'class': 'FileStream',
                'filename': client.daemon_log_file,
            },
            'env': get_env_with_venv_bin(),
        }]
    } # yapf: disable

    if not foreground:
        daemonize()

    arbiter = get_arbiter(**arbiter_config)
    pidfile = Pidfile(arbiter.pidfile)

    try:
        pidfile.create(os.getpid())
    except RuntimeError as exception:
        click.echo(str(exception))
        sys.exit(1)

    # Configure the logger
    loggerconfig = None
    loggerconfig = loggerconfig or arbiter.loggerconfig or None
    configure_logger(circus_logger, loglevel, logoutput, loggerconfig)

    # Main loop
    should_restart = True

    while should_restart:
        try:
            arbiter = arbiter
            future = arbiter.start()
            should_restart = False
            if check_future_exception_and_log(future) is None:
                should_restart = arbiter._restarting  # pylint: disable=protected-access
        except Exception as exception:
            # Emergency stop
            arbiter.loop.run_sync(arbiter._emergency_stop)  # pylint: disable=protected-access
            raise exception
        except KeyboardInterrupt:
            pass
        finally:
            arbiter = None
            if pidfile is not None:
                pidfile.unlink()

    sys.exit(0)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

from aiida.backends.testbase import AiidaTestCase
from aiida.common.exceptions import ModificationNotAllowed
from aiida.orm.calculation import Calculation


class TestCalcNode(AiidaTestCase):
    """
    These tests check the features of Calculation nodes that differ from the
    base Node type
    """
    boolval = True
    intval = 123
    floatval = 4.56
    stringval = 'aaaa'
    listval = [1, 's', True, None]
    dictval = {
        'num': 3, 'something': 'else', 'emptydict': {},
        'recursive': {
            'a': 1, 'b': True, 'c': 1.2, 'd': [1, 2, None],
            'e': {
                'z': 'z', 'x': None, 'xx': {}, 'yy': []
            }
        }
    }
    stateval = 'RUNNING'
    emptydict = {}
    emptylist = []

    def test_process_state(self):
        """
        Check the properties of a newly created bare Calculation
        """
        calculation = Calculation()

        self.assertEquals(calculation.is_terminated, False)
        self.assertEquals(calculation.is_excepted, False)
        self.assertEquals(calculation.is_killed, False)
        self.assertEquals(calculation.is_finished, False)
        self.assertEquals(calculation.is_finished_ok, False)
        self.assertEquals(calculation.is_failed, False)

    def test_calculation_updatable_attribute(self):
        """
        Check that updatable attributes and only those can be mutated for a stored but unsealed Calculation
        """
        a = Calculation()
        attrs_to_set = {
            'bool': self.boolval,
            'integer': self.intval,
            'float': self.floatval,
            'string': self.stringval,
            'dict': self.dictval,
            'list': self.listval,
            'state': self.stateval
        }

        for k, v in attrs_to_set.iteritems():
            a._set_attr(k, v)

        # Check before storing
        a._set_attr(Calculation.PROCESS_STATE_KEY, self.stateval)
        self.assertEquals(a.get_attr(Calculation.PROCESS_STATE_KEY), self.stateval)

        a.store()

        # Check after storing
        self.assertEquals(a.get_attr(Calculation.PROCESS_STATE_KEY), self.stateval)

        # I should be able to mutate the updatable attribute but not the others
        a._set_attr(Calculation.PROCESS_STATE_KEY, 'FINISHED')
        a._del_attr(Calculation.PROCESS_STATE_KEY)

        with self.assertRaises(ModificationNotAllowed):
            a._set_attr('bool', False)

        with self.assertRaises(ModificationNotAllowed):
            a._del_attr('bool')

        a.seal()

        # After sealing, even updatable attributes should be immutable
        with self.assertRaises(ModificationNotAllowed):
            a._set_attr(Calculation.PROCESS_STATE_KEY, 'FINISHED')

        with self.assertRaises(ModificationNotAllowed):
            a._del_attr(Calculation.PROCESS_STATE_KEY)


# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# Django settings for the AiiDA project.
import sys
import os

from aiida.common.exceptions import ConfigurationError, MissingConfigurationError

# get_property is used to read properties stored in the config json
from aiida.common.setup import (get_config, get_secret_key, get_property,
                                get_profile_config, parse_repository_uri)
from aiida.backends import settings
from aiida.utils.timezone import get_current_timezone

# Assumes that parent directory of aiida is root for
# things like templates/, SQL/ etc.  If not, change what follows...


AIIDA_DIR = os.path.dirname(os.path.abspath(__file__))
BASE_DIR = os.path.split(AIIDA_DIR)[0]
sys.path = [BASE_DIR] + sys.path

try:
    confs = get_config()
except MissingConfigurationError:
    raise MissingConfigurationError(
        "Please run the AiiDA Installation, no config found")

if settings.AIIDADB_PROFILE is None:
    raise ConfigurationError(
        "settings.AIIDADB_PROFILE not defined, did you load django"
        "through the AiiDA load_dbenv()?")

profile_conf = get_profile_config(settings.AIIDADB_PROFILE, conf_dict=confs)

# put all database specific portions of settings here
DBENGINE = profile_conf.get('AIIDADB_ENGINE', '')
DBNAME = profile_conf.get('AIIDADB_NAME', '')
DBUSER = profile_conf.get('AIIDADB_USER', '')
DBPASS = profile_conf.get('AIIDADB_PASS', '')
DBHOST = profile_conf.get('AIIDADB_HOST', '')
DBPORT = profile_conf.get('AIIDADB_PORT', '')
REPOSITORY_URI = profile_conf.get('AIIDADB_REPOSITORY_URI', '')

DATABASES = {
    'default': {
        # Add 'postgresql_psycopg2', 'mysql', or 'oracle'.
        'ENGINE': 'django.db.backends.' + DBENGINE,
        'NAME': DBNAME,
        'USER': DBUSER,
        'PASSWORD': DBPASS,
        'HOST': DBHOST,
        'PORT': DBPORT,
    }
}

# Checks on the REPOSITORY_* variables
try:
    REPOSITORY_URI
except NameError:
    raise ConfigurationError(
        "Please setup correctly the REPOSITORY_URI variable to "
        "a suitable directory on which you have write permissions.")

# Note: this variable might disappear in the future
REPOSITORY_PROTOCOL, REPOSITORY_PATH = parse_repository_uri(REPOSITORY_URI)

if settings.IN_RT_DOC_MODE:
    pass
elif REPOSITORY_PROTOCOL == 'file':
    if not os.path.isdir(REPOSITORY_PATH):
        try:
            # Try to create the local repository folders with needed parent
            # folders
            os.makedirs(REPOSITORY_PATH)
        except OSError:
            # Possibly here due to permission problems
            raise ConfigurationError(
                "Please setup correctly the REPOSITORY_PATH variable to "
                "a suitable directory on which you have write permissions. "
                "(I was not able to create the directory.)")
else:
    raise ConfigurationError("Only file protocol supported")

# CUSTOM USER CLASS
AUTH_USER_MODEL = 'db.DbUser'

# Make this unique, and don't share it with anybody.
# This is generated with the first run of 'verdi install'
SECRET_KEY = get_secret_key()

# Usual Django settings starts here.............

# Keep it to False! Otherwise every query is stored in memory
DEBUG = False
TEMPLATE_DEBUG = DEBUG

ADMINS = (
    # ('Your Name', 'your_email@example.com'),
)

MANAGERS = ADMINS

# Local time zone for this installation. Always choose the system timezone
TIME_ZONE = get_current_timezone().zone

# Language code for this installation. All choices can be found here:
# http://www.i18nguy.com/unicode/language-identifiers.html
LANGUAGE_CODE = 'en-us'

SITE_ID = 1

# If you set this to False, Django will make some optimizations so as not
# to load the internationalization machinery.
USE_I18N = False

# If you set this to False, Django will not format dates, numbers and
# calendars according to the current locale.
USE_L10N = False

# If you set this to False, Django will not use timezone-aware datetimes.
# For AiiDA, leave it as True, otherwise setting properties with dates will not work.
USE_TZ = True

# Absolute filesystem path to the directory that will hold user-uploaded files.
# Example: "/home/media/media.lawrence.com/media/"
MEDIA_ROOT = ''
# MEDIA_ROOT = '%s/templates/' % BASE_DIR

# URL that handles the media served from MEDIA_ROOT. Make sure to use a
# trailing slash.
# Examples: "http://media.lawrence.com/media/", "http://example.com/media/"
MEDIA_URL = ''

# Absolute path to the directory static files should be collected to.
# Don't put anything in this directory yourself; store your static files
# in apps' "static/" subdirectories and in STATICFILES_DIRS.
# Example: "/home/media/media.lawrence.com/static/"
STATIC_ROOT = None

# URL prefix for static files.
# Example: "http://media.lawrence.com/static/"
STATIC_URL = '/static/'

# Additional locations of static files
STATICFILES_DIRS = (
    # Put strings here, like "/home/html/static" or "C:/www/django/static".
    # Always use forward slashes, even on Windows.
    # Don't forget to use absolute paths, not relative paths.
)

# List of finder classes that know how to find static files in
# various locations.
STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
    #    'django.contrib.staticfiles.finders.DefaultStorageFinder',
)

# List of callables that know how to import templates from various sources.
TEMPLATE_LOADERS = (
    'django.template.loaders.filesystem.Loader',
    'django.template.loaders.app_directories.Loader',
    #     'django.template.loaders.eggs.Loader',
)

MIDDLEWARE_CLASSES = (
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    # Uncomment the next line for simple clickjacking protection:
    # 'django.middleware.clickjacking.XFrameOptionsMiddleware',
)

ROOT_URLCONF = 'aiida.backends.djsite.settings.urls'

# Python dotted path to the WSGI application used by Django's runserver.
WSGI_APPLICATION = 'aiida.backends.djsite.settings.wsgi.application'

TEMPLATE_DIRS = (
    # Put strings here, like "/home/html/django_templates" or "C:/www/django/templates".
    # Always use forward slashes, even on Windows.
    # Don't forget to use absolute paths, not relative paths.
)

TEMPLATE_CONTEXT_PROCESSORS = (
    "django.contrib.auth.context_processors.auth",
    "django.core.context_processors.debug",
    "django.core.context_processors.i18n",
    "django.core.context_processors.media",
    "django.core.context_processors.static",
    "django.core.context_processors.tz",
    "django.contrib.messages.context_processors.messages",
    "django.core.context_processors.request",
)

TEST_RUNNER = 'django.test.runner.DiscoverRunner'

INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.sites',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'aiida.backends.djsite.db',
]

# Automatic logging configuration for Django is disabled here
# and done for all backends in aiida/__init__.py
LOGGING_CONFIG = None

# -------------------------
# Tastypie (API) settings
# -------------------------
# For the time being, we support only json
TASTYPIE_DEFAULT_FORMATS = ['json']



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Classes for describing atomic orbitals.

Contains general Orbital class.
For subclasses of Orbital, see submodules.
"""
from aiida.common.exceptions import ValidationError, MissingPluginError
import math


class Orbital(object):
    """
    Base class for Orbitals. Can handle certain basic fields, their setting
    and validation. More complex Orbital objects should then inherit from
    this class

    :param position: the absolute position (three floats) units in angstrom
    :param x_orientation: x,y,z unit vector defining polar angle theta
                          in spherical coordinates unitless
    :param z_orientation: x,y,z unit vector defining azimuthal angle phi
                          in spherical coordinates unitless
    :param orientation_spin: x,y,z unit vector defining the spin orientation
                             unitless
    :param diffusivity: Float controls the radial term in orbital equation
                        units are reciprocal Angstrom.
    :param module_name: internal parameter, stores orbital type

    """
    #NOTE x_orientation, z_orientation, spin_orientation, diffusivity might
    #all need to be moved to RealHydrogenOrbital

    _base_fields = ('position',
                    'x_orientation',
                    'z_orientation',
                    'spin_orientation',
                    'diffusivity',
                    'module_name', # Actually, this one is system reserved
                    )

    def __init__(self):
        self._orbital_dict = {}

    def __repr__(self):
        module_name = self.get_orbital_dict()['module_name']
        return '<{}: {}>'.format(module_name, str(self))

    def __str__(self):
        raise NotImplementedError

    def _validate_keys(self, input_dict):
        """
        Checks all the input_dict and tries to validate them , to ensure
        that they have been properly set raises Exceptions indicating any
        problems that should arise during the validation

        :param input_dict: a dictionary of inputs
        :return: input_dict: the original dictionary with all validated kyes
                 now removed
        :return: validated_dict: a dictionary containing all the input keys
                 which have now been validated.
        """

        validated_dict = {}
        for k in self._base_fields:
            v = input_dict.pop(k, None)
            if k == "module_name":
                if v is None:
                    raise TypeError
                try:
                    OrbitalFactory(v)
                except (MissingPluginError, TypeError):
                    raise ValidationError("The module name {} was found to "
                                          "be invalid".format(v))
            if k == "position":
                if v is None:
                    validated_dict.update({k: v})
                    continue
                try:
                    v = list(float(i) for i in v)
                    if len(v) != 3:
                        raise ValueError
                except (ValueError, TypeError):
                        raise ValueError("Wrong format for position, must be a"
                                         " list of three float numbers.")
            if "orientation" in k :
                if v is None:
                    validated_dict.update({k: v})
                    continue
                try:
                    v = list(float(i) for i in v)
                    if len(v) != 3:
                        raise ValueError
                except (ValueError, TypeError):
                        raise ValueError("Wrong format for {}, must be a"
                                         " list of three float numbers.")
            # From a spherical cooridnate version of orientation
            # try:
            #     v = tuple(float(i) for i in v)
            #     if len(v) != (2):
            #         raise ValueError
            #     if v[0] >= 2*math.pi or v[0] <= 0:
            #         raise ValueError
            #     if v[1] >= math.pi or v[1] <= 0:
            #         raise ValueError
            # except(ValueError, TypeError):
            #     raise ValueError("Wrong format for {}, must be two tuples"
            #                      " each having two floats theta, phi where"
            #                      " 0<=theta<2pi and 0<=phi<=pi.".format(k))
            if k == "diffusivity":
                if v is None:
                    validated_dict.update({k: v})
                    continue
                try:
                    v = float(v)
                except ValueError:
                    raise ValidationError("Diffusivity must always be a float")
            validated_dict.update({k: v})
        return validated_dict

    def set_orbital_dict(self, init_dict):
        """
        Sets the orbital_dict, which can vary depending on the particular
        implementation of this base class.

        :param init_dict: the initialization dictionary
        """
        if not isinstance(init_dict, dict):
            raise Exception('You must supply a dict as an init')
        # Adds the module_name in hard-coded manner
        init_dict.update({"module_name": self._get_module_name()})
        validated_dict = self._validate_keys(init_dict)
        for k, v in validated_dict.iteritems():
            self._orbital_dict[k] = v

    def get_orbital_dict(self):
        """
        returns the internal keys as a dictionary
        """
        output = {}
        for k in self._default_fields:
            try:
                output[k] = self._orbital_dict[k]
            except KeyError:
                pass
        return output

    def _get_module_name(self):
        """
        Sets the module name, or label, to the orbital
        """
        return self.__module__.split('.')[-1]


def OrbitalFactory(entry_point):
    """
    Return the Orbital plugin class for a given entry point

    :param entry_point: the entry point name of the Orbital plugin
    """
    return BaseFactory('aiida.common.orbital', entry_point)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from pytz import UTC
from sqlalchemy import Column
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm.attributes import flag_modified
from sqlalchemy.schema import UniqueConstraint
from sqlalchemy.types import Integer, String, DateTime

import aiida.backends.sqlalchemy as sa
from aiida.backends.sqlalchemy.models.base import Base
from aiida.utils import timezone




class DbSetting(Base):
    __tablename__ = "db_dbsetting"
    __table_args__ = (UniqueConstraint('key'),)
    id = Column(Integer, primary_key=True)

    key = Column(String(255), index=True, nullable=False)
    val = Column(JSONB, default={})

    # I also add a description field for the variables
    description = Column(String(255), default='', nullable=False)
    time = Column(DateTime(timezone=True), default=timezone.utc)

    def __str__(self):
        return "'{}'={}".format(self.key, self.getvalue())
    
    @classmethod
    def set_value(cls, key, value, with_transaction=True,
                  subspecifier_value=None, other_attribs={},
                  stop_if_existing=False):

        setting = sa.get_scoped_session().query(DbSetting).filter_by(key=key).first()
        if setting is not None:
            if stop_if_existing:
                return
        else:
            setting = cls()

        setting.key = key
        setting.val = value
        flag_modified(setting, "val")
        setting.time = timezone.datetime.now(tz=UTC)
        if "description" in other_attribs.keys():
            setting.description = other_attribs["description"]
        setting.save()

    def getvalue(self):
        """
        This can be called on a given row and will get the corresponding value.
        """
        return self.val

    def get_description(self):
        """
        This can be called on a given row and will get the corresponding
        description.
        """
        return self.description

    @classmethod
    def del_value(cls, key, only_children=False, subspecifier_value=None):
        setting = sa.get_scoped_session().query(DbSetting).filter(key=key)
        setting.val = None
        setting.time = timezone.datetime.utcnow()
        flag_modified(setting, "val")
        setting.save()
